# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Storage of test results.

A Repository provides storage and indexing of results.

The AbstractRepository class defines the contract to which any Repository
implementation must adhere.

The stestr.repository.file module (see: :ref:`api_repository_file` is the usual
repository that will be used. The stestr.repository.memory module (see:
:ref:`api_repository_memory`) provides a memory only repository useful for
internal testing.

Repositories are identified by their URL, and new ones are made by calling
the initialize function in the appropriate repository module.
"""

from testtools import StreamToDict


class AbstractRepositoryFactory(object):
    """Interface for making or opening repositories."""

    def initialise(self, url):
        """Create a repository at URL.

        Call on the class of the repository you wish to create.
        """
        raise NotImplementedError(self.initialise)

    def open(self, url):
        """Open the repository at url.

        Raise RepositoryNotFound if there is no repository at the given url.
        """
        raise NotImplementedError(self.open)


class AbstractRepository(object):
    """The base class for Repository implementations.

    There are no interesting attributes or methods as yet.
    """

    def count(self):
        """Return the number of test runs this repository has stored.

        :return count: The count of test runs stored in the repository.
        """
        raise NotImplementedError(self.count)

    def get_failing(self):
        """Get a TestRun that contains all of and only current failing tests.

        :return: a TestRun.
        """
        raise NotImplementedError(self.get_failing)

    def get_inserter(self, partial=False, run_id=None):
        """Get an inserter that will insert a test run into the repository.

        Repository implementations should implement _get_inserter.

        get_inserter() does not add timing data to streams: it should be
        provided by the caller of get_inserter (e.g. commands.load).

        :param partial: DEPREACTED: If True, the stream being inserted only
            executed some tests rather than all the projects tests. This
            option is deprecated and no longer does anything. It will be
            removed in the future.
        :return an inserter: Inserters meet the extended TestResult protocol
            that testtools 0.9.2 and above offer. The startTestRun and
            stopTestRun methods in particular must be called.
        """
        return self._get_inserter(partial, run_id)

    def _get_inserter(self, partial=False, run_id=None):
        """Get an inserter for get_inserter.

        The result is decorated with an AutoTimingTestResultDecorator.
        """
        raise NotImplementedError(self._get_inserter)

    def get_latest_run(self):
        """Return the latest run.

        Equivalent to get_test_run(latest_id()).
        """
        return self.get_test_run(self.latest_id())

    def get_test_run(self, run_id):
        """Retrieve a TestRun object for run_id.

        :param run_id: The test run id to retrieve.
        :return: A TestRun object.
        """
        raise NotImplementedError(self.get_test_run)

    def get_test_times(self, test_ids):
        """Retrieve estimated times for the tests test_ids.

        :param test_ids: The test ids to query for timing data.
        :return: A dict with two keys: 'known' and 'unknown'. The unknown
            key contains a set with the test ids that did run. The known
            key contains a dict mapping test ids to time in seconds.
        """
        test_ids = frozenset(test_ids)
        known_times = self._get_test_times(test_ids)
        unknown_times = test_ids - set(known_times)
        return dict(known=known_times, unknown=unknown_times)

    def _get_test_times(self, test_ids):
        """Retrieve estimated times for tests test_ids.

        :param test_ids: The test ids to query for timing data.
        :return: A dict mapping test ids to duration in seconds. Tests that no
            timing data is present for should not be returned - the base class
            get_test_times function will collate the missing test ids and put
            that in to its result automatically.
        """
        raise NotImplementedError(self._get_test_times)

    def latest_id(self):
        """Return the run id for the most recently inserted test run."""
        raise NotImplementedError(self.latest_id)

    def get_test_ids(self, run_id):
        """Return the test ids from the specified run.

        :param run_id: the id of the test run to query.
        :return: a list of test ids for the tests that
            were part of the specified test run.
        """
        run = self.get_test_run(run_id)
        ids = []

        def gather(test_dict):
            ids.append(test_dict['id'])

        result = StreamToDict(gather)
        result.startTestRun()
        try:
            run.get_test().run(result)
        finally:
            result.stopTestRun()
        return ids


class AbstractTestRun(object):
    """A test run that has been stored in a repository.

    Should implement the StreamResult protocol as well
    as the stestr specific methods documented here.
    """

    def get_id(self):
        """Get the id of the test run.

        Sometimes test runs will not have an id, e.g. test runs for
        'failing'. In that case, this should return None.
        """
        raise NotImplementedError(self.get_id)

    def get_subunit_stream(self):
        """Get a subunit stream for this test run."""
        raise NotImplementedError(self.get_subunit_stream)

    def get_test(self):
        """Get a testtools.TestCase-like object that can be run.

        :return: A TestCase like object which can be run to get the individual
            tests reported to a testtools.StreamResult/TestResult.
            (Clients of repository should provide an ExtendedToStreamDecorator
            decorator to permit either API to be used).
        """
        raise NotImplementedError(self.get_test)


class RepositoryNotFound(Exception):
    """Raised when we try to open a repository that isn't there."""

    def __init__(self, url):
        self.url = url
        msg = 'No repository found in %s. Create one by running "stestr init".'
        Exception.__init__(self, msg % url)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from stestr.commands import slowest
from stestr.tests import base


class TestSlowest(base.TestCase):
    def test_format_times(self):
        times = [('test_id_a', 12.34), ('test_id_b', 1.34)]
        res = slowest.format_times(times)
        self.assertEqual([('test_id_a', '12.340'), ('test_id_b', ' 1.340')],
                         res)

    def test_format_times_with_zero(self):
        times = [('test_id_a', 0), ('test_id_b', 1.34)]
        res = slowest.format_times(times)
        self.assertEqual([('test_id_a', '0.000'), ('test_id_b', '1.340')],
                         res)

    def test_format_times_all_zero(self):
        times = [('test_id_a', 0), ('test_id_b', 0.00)]
        res = slowest.format_times(times)
        self.assertEqual([('test_id_a', '0.000'), ('test_id_b', '0.000')],
                         res)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import io
import os
import shutil
import subprocess
import tempfile

import six
from six import StringIO
import subunit as subunit_lib
import testtools
import yaml

from stestr.tests import base


class TestReturnCodes(base.TestCase):
    def setUp(self):
        super(TestReturnCodes, self).setUp()
        # Setup test dirs
        self.directory = tempfile.mkdtemp(prefix='stestr-unit')
        self.addCleanup(shutil.rmtree, self.directory)
        self.test_dir = os.path.join(self.directory, 'tests')
        os.mkdir(self.test_dir)
        # Setup Test files
        self.testr_conf_file = os.path.join(self.directory, '.stestr.conf')
        self.setup_cfg_file = os.path.join(self.directory, 'setup.cfg')
        self.passing_file = os.path.join(self.test_dir, 'test_passing.py')
        self.failing_file = os.path.join(self.test_dir, 'test_failing.py')
        self.init_file = os.path.join(self.test_dir, '__init__.py')
        self.setup_py = os.path.join(self.directory, 'setup.py')
        shutil.copy('stestr/tests/files/testr-conf', self.testr_conf_file)
        shutil.copy('stestr/tests/files/passing-tests', self.passing_file)
        shutil.copy('stestr/tests/files/failing-tests', self.failing_file)
        shutil.copy('setup.py', self.setup_py)
        shutil.copy('stestr/tests/files/setup.cfg', self.setup_cfg_file)
        shutil.copy('stestr/tests/files/__init__.py', self.init_file)

        self.stdout = StringIO()
        self.stderr = StringIO()
        # Change directory, run wrapper and check result
        self.addCleanup(os.chdir, os.path.abspath(os.curdir))
        os.chdir(self.directory)
        subprocess.call('stestr init', shell=True)

    def _get_cmd_stdout(self, cmd):
        p = subprocess.Popen(cmd, shell=True,
                             stdout=subprocess.PIPE)
        out = p.communicate()
        self.assertEqual(0, p.returncode)
        return out

    def assertRunExit(self, cmd, expected, subunit=False, stdin=None):
        if stdin:
            p = subprocess.Popen(
                "%s" % cmd, shell=True, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate(stdin)
        else:
            p = subprocess.Popen(
                "%s" % cmd, shell=True,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()

        if not subunit:
            self.assertEqual(
                p.returncode, expected,
                "Stdout: %s; Stderr: %s" % (out, err))
            return (out, err)
        else:
            self.assertEqual(p.returncode, expected,
                             "Expected return code: %s doesn't match actual "
                             "return code of: %s" % (expected, p.returncode))
            output_stream = io.BytesIO(out)
            stream = subunit_lib.ByteStreamToStreamResult(output_stream)
            starts = testtools.StreamResult()
            summary = testtools.StreamSummary()
            tests = []

            def _add_dict(test):
                tests.append(test)

            outcomes = testtools.StreamToDict(functools.partial(_add_dict))
            result = testtools.CopyStreamResult([starts, outcomes, summary])
            result.startTestRun()
            try:
                stream.run(result)
            finally:
                result.stopTestRun()
            self.assertThat(len(tests), testtools.matchers.GreaterThan(0))
            return (out, err)

    def test_empty_config_file_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        self.assertRunExit(
            'stestr --user-config=%s run' % path, 1)

    def test_empty_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        self.assertRunExit(
            'stestr --user-config=%s run passing' % path, 0)

    def test_no_subunit_trace_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'no-subunit-trace': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        out, err = self.assertRunExit(
            'stestr --user-config=%s run passing' % path, 0)
        out = six.text_type(out)
        self.assertIn('PASSED (id=0)', out)
        self.assertNotIn('Totals', out)
        self.assertNotIn('Worker Balance', out)
        self.assertNotIn('Sum of execute time for each test:', out)
        self.assertNotIn('Runtime (s)', out)

    def test_no_subunit_trace_config_file_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'no-subunit-trace': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        out, err = self.assertRunExit(
            'stestr --user-config=%s run' % path, 1)
        out = six.text_type(out)
        self.assertIn('FAILED (id=0, failures=2)', out)
        self.assertNotIn('Totals', out)
        self.assertNotIn('Worker Balance', out)
        self.assertNotIn('Sum of execute time for each test:', out)

    def test_no_subunit_trace_config_file_force_subunit_trace(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'no-subunit-trace': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        out, err = self.assertRunExit(
            'stestr --user-config=%s run --force-subunit-trace passing' % path,
            0)
        out = six.text_type(out)
        self.assertNotIn('PASSED (id=0)', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    def test_abbreviate_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'abbreviate': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        out, err = self.assertRunExit(
            'stestr --user-config=%s run passing' % path, 0)
        out = six.text_type(out)
        self.assertIn('..', out)
        self.assertNotIn('PASSED (id=0)', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    def test_abbreviate_config_file_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'abbreviate': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        # NOTE(mtreinish): Running serially here to ensure a consistent
        # execution order for confirming the abbreviated output.
        out, err = self.assertRunExit(
            'stestr --user-config=%s run --serial' % path, 1)
        out = six.text_type(out)
        self.assertIn('FF..', out)
        self.assertNotIn('FAILED (id=0, failures=2)', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    def test_no_subunit_trace_slowest_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'no-subunit-trace': True,
                    'slowest': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        out, err = self.assertRunExit(
            'stestr --user-config=%s run passing' % path, 0)
        out = six.text_type(out)
        self.assertIn('PASSED (id=0)', out)
        self.assertNotIn('Totals', out)
        self.assertNotIn('Worker Balance', out)
        self.assertNotIn('Sum of execute time for each test:', out)
        self.assertIn('Runtime (s)', out)

    def test_failing_list_config_file(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'no-subunit-trace': True,
                    'slowest': True,
                },
                'failing': {
                    'list': True
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        self.assertRunExit('stestr --user-config=%s run' % path, 1)
        out, err = self.assertRunExit('stestr --user-config=%s failing' % path,
                                      1)
        out = six.text_type(out)
        self.assertNotIn('FAILED (id=0, failures=2)', out)
        self.assertNotIn('FAIL:', out)
        self.assertIn('tests.test_failing.FakeTestClass.test_pass', out)
        self.assertIn('tests.test_failing.FakeTestClass.test_pass_list', out)

    def test_no_subunit_trace_last_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'slowest': True,
                },
                'failing': {
                    'list': True
                },
                'last': {
                    'no-subunit-trace': True,
                },
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        run_out, run_err = self.assertRunExit(
            'stestr --user-config=%s run passing' % path, 0)
        out, err = self.assertRunExit('stestr --user-config=%s last' % path,
                                      0)
        run_out = six.text_type(run_out)
        out = six.text_type(out)
        self.assertIn('PASSED (id=0)', out)
        self.assertNotIn('Totals', out)
        self.assertNotIn('Worker Balance', out)
        self.assertNotIn('Sum of execute time for each test:', out)
        self.assertNotIn('Runtime (s)', out)
        self.assertIn('Totals', run_out)
        self.assertIn('Worker Balance', run_out)
        self.assertIn('Sum of execute time for each test:', run_out)
        self.assertIn('Runtime (s)', run_out)

    def test_subunit_trace_load_from_config_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'slowest': True,
                },
                'failing': {
                    'list': True
                },
                'last': {
                    'no-subunit-trace': True,
                },
                'load': {
                    'subunit-trace': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))

        self.assertRunExit('stestr --user-config=%s run passing' % path, 0)
        stream = self._get_cmd_stdout(
            'stestr --user-config=%s last --subunit' % path)[0]
        out, err = self.assertRunExit('stestr --user-config=%s load' % path,
                                      0, stdin=stream)
        out = six.text_type(out)
        self.assertNotIn('PASSED (id=0)', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    def test_subunit_trace_load_from_config_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'slowest': True,
                },
                'failing': {
                    'list': True
                },
                'last': {
                    'no-subunit-trace': True,
                },
                'load': {
                    'subunit-trace': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))

        self.assertRunExit('stestr --user-config=%s run' % path, 1)
        stream = self._get_cmd_stdout(
            'stestr --user-config=%s last --subunit' % path)[0]
        out, err = self.assertRunExit('stestr --user-config=%s load' % path,
                                      0, stdin=stream)
        out = six.text_type(out)
        self.assertNotIn('FAILED (id=0, failures=2)', out)
        self.assertNotIn('FF..', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    @testtools.skip('Abbreviated output not displaying')
    def test_abbreviate_load_from_config_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'slowest': True,
                },
                'failing': {
                    'list': True
                },
                'last': {
                    'no-subunit-trace': True,
                },
                'load': {
                    'abbreviate': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))

        self.assertRunExit('stestr --user-config=%s run passing' % path, 0)
        stream = self._get_cmd_stdout(
            'stestr --user-config=%s last --subunit' % path)[0]
        out, err = self.assertRunExit('stestr --user-config=%s load' % path,
                                      0, stdin=stream)
        out = six.text_type(out)
        self.assertNotIn('PASSED (id=0)', out)
        self.assertIn('..', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    @testtools.skip('Abbreviated output not displaying')
    def test_abbreviate_load_from_config_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, 'wb', 0)
        self.addCleanup(conf_file.close)
        contents = six.text_type(
            yaml.dump({
                'run': {
                    'slowest': True,
                },
                'failing': {
                    'list': True
                },
                'last': {
                    'no-subunit-trace': True,
                },
                'load': {
                    'abbreviate': True,
                }
            }, default_flow_style=False))
        conf_file.write(contents.encode('utf-8'))
        # NOTE(mtreinish): Running serially here to ensure a consistent
        # execution order for confirming the abbreviated output.
        self.assertRunExit('stestr --user-config=%s run --serial' % path, 1)
        stream = self._get_cmd_stdout(
            'stestr --user-config=%s last --subunit' % path)[0]
        out, err = self.assertRunExit('stestr --user-config=%s load' % path,
                                      0, stdin=stream)
        out = six.text_type(out)
        self.assertNotIn('FAILED (id=0, failures=2)', out)
        self.assertIn('FF..', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

import mock
import six

from stestr import selection
from stestr.tests import base


class TestSelection(base.TestCase):
    def test_filter_tests_no_filter(self):
        test_list = ['a', 'b', 'c']
        result = selection.filter_tests(None, test_list)
        self.assertEqual(test_list, result)

    def test_filter_tests(self):
        test_list = ['a', 'b', 'c']
        result = selection.filter_tests(['a'], test_list)
        self.assertEqual(['a'], result)

    def test_filter_invalid_regex(self):
        test_list = ['a', 'b', 'c']
        with mock.patch('sys.exit', side_effect=ImportError) as mock_exit:
            self.assertRaises(ImportError, selection.filter_tests,
                              ['fake_regex_with_bad_part[The-BAD-part]'],
                              test_list)
            mock_exit.assert_called_once_with(5)


class TestBlackReader(base.TestCase):
    def test_black_reader(self):
        blacklist_file = six.StringIO()
        for i in range(4):
            blacklist_file.write('fake_regex_%s\n' % i)
            blacklist_file.write('fake_regex_with_note_%s # note\n' % i)
        blacklist_file.seek(0)
        with mock.patch('six.moves.builtins.open',
                        return_value=blacklist_file):
            result = selection.black_reader('fake_path')
        self.assertEqual(2 * 4, len(result))
        note_cnt = 0
        # not assuming ordering, mainly just testing the type
        for r in result:
            self.assertEqual(r[2], [])
            if r[1] == 'note':
                note_cnt += 1
            self.assertIn('search', dir(r[0]))  # like a compiled regexp
        self.assertEqual(note_cnt, 4)

    def test_invalid_regex(self):
        blacklist_file = six.StringIO()
        blacklist_file.write("fake_regex_with_bad_part[The-BAD-part]")
        blacklist_file.seek(0)
        with mock.patch('six.moves.builtins.open',
                        return_value=blacklist_file):
            with mock.patch('sys.exit') as mock_exit:
                selection.black_reader('fake_path')
                mock_exit.assert_called_once_with(5)


class TestConstructList(base.TestCase):
    def test_simple_re(self):
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])']
        result = selection.construct_list(test_lists, regexes=['foo'])
        self.assertEqual(list(result), ['fake_test(scen)[egg,foo])'])

    def test_simple_black_re(self):
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])']
        result = selection.construct_list(test_lists, black_regex='foo')
        self.assertEqual(list(result), ['fake_test(scen)[tag,bar])'])

    def test_invalid_black_re(self):
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])']
        invalid_regex = "fake_regex_with_bad_part[The-BAD-part]"
        with mock.patch('sys.exit', side_effect=ImportError) as exit_mock:
            self.assertRaises(ImportError, selection.construct_list,
                              test_lists, black_regex=invalid_regex)
            exit_mock.assert_called_once_with(5)

    def test_blacklist(self):
        black_list = [(re.compile('foo'), 'foo not liked', [])]
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])']
        with mock.patch('stestr.selection.black_reader',
                        return_value=black_list):
            result = selection.construct_list(test_lists,
                                              blacklist_file='file',
                                              regexes=['fake_test'])
        self.assertEqual(list(result), ['fake_test(scen)[tag,bar])'])

    def test_whitelist(self):
        white_list = [re.compile('fake_test1'), re.compile('fake_test2')]
        test_lists = ['fake_test1[tg]', 'fake_test2[tg]', 'fake_test3[tg]']
        white_getter = 'stestr.selection._get_regex_from_whitelist_file'
        with mock.patch(white_getter,
                        return_value=white_list):
            result = selection.construct_list(test_lists,
                                              whitelist_file='file')
        self.assertEqual(set(result),
                         set(('fake_test1[tg]', 'fake_test2[tg]')))

    def test_whitelist_invalid_regex(self):
        whitelist_file = six.StringIO()
        whitelist_file.write("fake_regex_with_bad_part[The-BAD-part]")
        whitelist_file.seek(0)
        with mock.patch('six.moves.builtins.open',
                        return_value=whitelist_file):
            with mock.patch('sys.exit') as mock_exit:
                selection._get_regex_from_whitelist_file('fake_path')
                mock_exit.assert_called_once_with(5)

    def test_whitelist_blacklist_re(self):
        white_list = [re.compile('fake_test1'), re.compile('fake_test2')]
        test_lists = ['fake_test1[tg]', 'fake_test2[spam]',
                      'fake_test3[tg,foo]', 'fake_test4[spam]']
        black_list = [(re.compile('spam'), 'spam not liked', [])]
        white_getter = 'stestr.selection._get_regex_from_whitelist_file'
        with mock.patch(white_getter,
                        return_value=white_list):
            with mock.patch('stestr.selection.black_reader',
                            return_value=black_list):
                result = selection.construct_list(test_lists, 'black_file',
                                                  'white_file', ['foo'])
        self.assertEqual(set(result),
                         set(('fake_test1[tg]', 'fake_test3[tg,foo]')))

    def test_overlapping_black_regex(self):

        black_list = [(re.compile('compute.test_keypairs.KeypairsTestV210'),
                       '', []),
                      (re.compile('compute.test_keypairs.KeypairsTestV21'),
                       '', [])]
        test_lists = [
            'compute.test_keypairs.KeypairsTestV210.test_create_keypair',
            'compute.test_keypairs.KeypairsTestV21.test_create_keypair',
            'compute.test_fake.FakeTest.test_fake_test']
        with mock.patch('stestr.selection.black_reader',
                        return_value=black_list):
            result = selection.construct_list(test_lists,
                                              blacklist_file='file',
                                              regexes=['fake_test'])
        self.assertEqual(
            list(result), ['compute.test_fake.FakeTest.test_fake_test'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil
import subprocess
import tempfile

import six

from stestr.tests import base


class TestBisectReturnCodes(base.TestCase):
    def setUp(self):
        super(TestBisectReturnCodes, self).setUp()
        # Setup test dirs
        self.directory = tempfile.mkdtemp(prefix='stestr-unit')
        self.addCleanup(shutil.rmtree, self.directory)
        self.test_dir = os.path.join(self.directory, 'tests')
        os.mkdir(self.test_dir)
        # Setup Test files
        self.testr_conf_file = os.path.join(self.directory, '.stestr.conf')
        self.setup_cfg_file = os.path.join(self.directory, 'setup.cfg')
        self.init_file = os.path.join(self.test_dir, '__init__.py')
        self.setup_py = os.path.join(self.directory, 'setup.py')
        self.user_config = os.path.join(self.directory, 'stestr.yaml')
        shutil.copy('stestr/tests/files/testr-conf', self.testr_conf_file)
        shutil.copy('setup.py', self.setup_py)
        shutil.copy('stestr/tests/files/setup.cfg', self.setup_cfg_file)
        shutil.copy('stestr/tests/files/__init__.py', self.init_file)
        shutil.copy('stestr/tests/files/stestr.yaml', self.user_config)

        # Move around the test code
        self.serial_fail_file = os.path.join(self.test_dir,
                                             'test_serial_fails.py')
        shutil.copy('stestr/tests/files/bisect-fail-serial-tests',
                    self.serial_fail_file)

        # Change directory, run wrapper and check result
        self.addCleanup(os.chdir, os.path.abspath(os.curdir))
        os.chdir(self.directory)
        subprocess.call('stestr init', shell=True)

    def test_bisect_serial_fail_detected(self):
        p = subprocess.Popen(
            "stestr run --serial", shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(1, p.returncode,
                         'stestr run returned an unexpected return code'
                         'Stdout: %s\nStderr: %s' % (out, err))
        p_analyze = subprocess.Popen(
            "stestr --user-config stestr.yaml run --analyze-isolation",
            shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p_analyze.communicate()
        out = out.decode('utf-8')
        # For debugging potential failures
        lines = six.text_type(out.rstrip()).splitlines()
        self.assertEqual(3, p_analyze.returncode,
                         'Analyze isolation returned an unexpected return code'
                         '\nStdout: %s\nStderr: %s' % (out, err))
        last_line = ('tests.test_serial_fails.TestFakeClass.test_B  '
                     'tests.test_serial_fails.TestFakeClass.test_A')
        self.assertEqual(last_line, lines[-1])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import subprocess

import mock

from stestr import test_processor
from stestr.tests import base


class TestTestProcessorFixture(base.TestCase):

    def setUp(self):
        super(TestTestProcessorFixture, self).setUp()
        self._fixture = test_processor.TestProcessorFixture(
            mock.sentinel.test_ids, mock.sentinel.options,
            mock.sentinel.cmd_template, mock.sentinel.listopt,
            mock.sentinel.idoption, mock.sentinel.repository)

    @mock.patch.object(subprocess, 'Popen')
    @mock.patch.object(test_processor, 'sys')
    def _check_start_process(self, mock_sys, mock_Popen, platform='win32',
                             expected_fn=None):
        mock_sys.platform = platform

        self._fixture._start_process(mock.sentinel.cmd)

        mock_Popen.assert_called_once_with(
            mock.sentinel.cmd, shell=True, stdout=subprocess.PIPE,
            stdin=subprocess.PIPE, preexec_fn=expected_fn)

    def test_start_process_win32(self):
        self._check_start_process()

    def test_start_process_linux(self):
        self._check_start_process(
            platform='linux2', expected_fn=self._fixture._clear_SIGPIPE)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Load data into a repository."""


import functools
import os
import sys
import warnings

from cliff import command
import subunit
import testtools

from stestr import output
from stestr.repository import abstract as repository
from stestr.repository import util
from stestr import results
from stestr import subunit_trace
from stestr import user_config
from stestr import utils


class Load(command.Command):
    """Load a subunit stream into a repository.

    Failing tests are shown on the console and a summary of the stream
    is printed at the end.

    Without --subunit, the process exit code will be non-zero if the test
    run was not successful. With --subunit, the process exit code is
    non-zero if the subunit stream could not be generated successfully.
    The test results and run status are included in the subunit stream, so
    the stream should be used to determining the result of the run instead
    of the exit code when using the --subunit flag.
    """

    def get_parser(self, prog_name):
        parser = super(Load, self).get_parser(prog_name)
        parser.add_argument("files", nargs="*", default=False,
                            help="The subunit v2 stream files to load into the"
                            " repository")
        parser.add_argument("--partial", action="store_true",
                            default=False,
                            help="DEPRECATED: The stream being loaded was a "
                            "partial run. This option is deprecated and no "
                            "does anything. It will be removed in the future")
        parser.add_argument("--force-init", action="store_true",
                            default=False,
                            help="Initialise the repository if it does not "
                            "exist already")
        parser.add_argument("--subunit", action="store_true",
                            default=False,
                            help="Display results in subunit format.")
        parser.add_argument("--id", "-i", default=None,
                            help="Append the stream into an existing entry in "
                            "the repository")
        parser.add_argument("--subunit-trace", action='store_true',
                            default=False,
                            help="Display the loaded stream through the "
                            "subunit-trace output filter")
        parser.add_argument('--color', action='store_true', default=False,
                            help='Enable color output in the subunit-trace '
                            'output, if subunit-trace output is enabled. If '
                            'subunit-trace is disable this does nothing.')
        parser.add_argument('--abbreviate', action='store_true',
                            dest='abbreviate',
                            help='Print one character status for each test')
        parser.add_argument('--suppress-attachments', action='store_true',
                            dest='suppress_attachments',
                            help='If set do not print stdout or stderr '
                            'attachment contents on a successful test '
                            'execution')
        return parser

    def take_action(self, parsed_args):
        user_conf = user_config.get_user_config(self.app_args.user_config)
        args = parsed_args
        if getattr(user_conf, 'load', False):
            force_init = args.force_init or user_conf.load.get('force-init',
                                                               False)
            pretty_out = args.subunit_trace or user_conf.load.get(
                'subunit-trace', False)
            color = args.color or user_conf.load.get('color', False)
            abbreviate = args.abbreviate or user_conf.load.get('abbreviate',
                                                               False)
            suppress_attachments = (
                args.suppress_attachments or user_conf.load.get(
                    'suppress-attachments', False))
        else:
            force_init = args.force_init
            pretty_out = args.subunit_trace
            color = args.color
            abbreviate = args.abbreviate
            suppress_attachments = args.suppress_attachments
        verbose_level = self.app.options.verbose_level
        stdout = open(os.devnull, 'w') if verbose_level == 0 else sys.stdout
        load(repo_type=self.app_args.repo_type,
             repo_url=self.app_args.repo_url,
             partial=args.partial, subunit_out=args.subunit,
             force_init=force_init, streams=args.files,
             pretty_out=pretty_out, color=color,
             stdout=stdout, abbreviate=abbreviate,
             suppress_attachments=suppress_attachments, serial=True)


def load(force_init=False, in_streams=None,
         partial=False, subunit_out=False, repo_type='file', repo_url=None,
         run_id=None, streams=None, pretty_out=False, color=False,
         stdout=sys.stdout, abbreviate=False, suppress_attachments=False,
         serial=False):
    """Load subunit streams into a repository

    This function will load subunit streams into the repository. It will
    output to STDOUT the results from the input stream. Internally this is
    used by the run command to both output the results as well as store the
    result in the repository.

    :param bool force_init: Initialize the specifiedrepository if it hasn't
        been created.
    :param list in_streams: A list of file objects that will be saved into the
        repository
    :param bool partial: DEPRECATED: Specify the input is a partial stream.
        This option is deprecated and no longer does anything. It will be
        removed in the future.
    :param bool subunit_out: Output the subunit stream to stdout
    :param str repo_type: This is the type of repository to use. Valid choices
        are 'file' and 'sql'.
    :param str repo_url: The url of the repository to use.
    :param run_id: The optional run id to save the subunit stream to.
    :param list streams: A list of file paths to read for the input streams.
    :param bool pretty_out: Use the subunit-trace output filter for the loaded
        stream.
    :param bool color: Enabled colorized subunit-trace output
    :param file stdout: The output file to write all output to. By default
        this is sys.stdout
    :param bool abbreviate: Use abbreviated output if set true
    :param bool suppress_attachments: When set true attachments subunit_trace
        will not print attachments on successful test execution.

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """
    if partial:
        warnings.warn('The partial flag is deprecated and has no effect '
                      'anymore')

    try:
        repo = util.get_repo_open(repo_type, repo_url)
    except repository.RepositoryNotFound:
        if force_init:
            repo = util.get_repo_initialise(repo_type, repo_url)
        else:
            raise
    # Not a full implementation of TestCase, but we only need to iterate
    # back to it. Needs to be a callable - its a head fake for
    # testsuite.add.
    if in_streams:
        streams = utils.iter_streams(in_streams, 'subunit')
    elif streams:
        opener = functools.partial(open, mode='rb')
        streams = map(opener, streams)
    else:
        streams = [sys.stdin]

    def mktagger(pos, result):
        return testtools.StreamTagger([result], add=['worker-%d' % pos])

    def make_tests():
        for pos, stream in enumerate(streams):
            # Calls StreamResult API.
            case = subunit.ByteStreamToStreamResult(
                stream, non_subunit_name='stdout')
            decorate = functools.partial(mktagger, pos)
            case = testtools.DecorateTestCaseResult(case, decorate)
            yield (case, str(pos))
    if not run_id:
        inserter = repo.get_inserter()
    else:
        inserter = repo.get_inserter(run_id=run_id)

    retval = 0
    if serial:
        for stream in streams:
            # Calls StreamResult API.
            case = subunit.ByteStreamToStreamResult(
                stream, non_subunit_name='stdout')
            result = _load_case(inserter, repo, case, subunit_out, pretty_out,
                                color, stdout, abbreviate,
                                suppress_attachments)
            if result or retval:
                retval = 1
            else:
                retval = 0
    else:
        case = testtools.ConcurrentStreamTestSuite(make_tests)
        retval = _load_case(inserter, repo, case, subunit_out, pretty_out,
                            color, stdout, abbreviate, suppress_attachments)

    return retval


def _load_case(inserter, repo, case, subunit_out, pretty_out,
               color, stdout, abbreviate, suppress_attachments):
    if subunit_out:
        output_result, summary_result = output.make_result(inserter.get_id,
                                                           output=stdout)
    elif pretty_out:
        outcomes = testtools.StreamToDict(
            functools.partial(subunit_trace.show_outcome, stdout,
                              enable_color=color, abbreviate=abbreviate,
                              suppress_attachments=suppress_attachments))
        summary_result = testtools.StreamSummary()
        output_result = testtools.CopyStreamResult([outcomes, summary_result])
        output_result = testtools.StreamResultRouter(output_result)
        cat = subunit.test_results.CatFiles(stdout)
        output_result.add_rule(cat, 'test_id', test_id=None)
    else:
        try:
            previous_run = repo.get_latest_run()
        except KeyError:
            previous_run = None
        output_result = results.CLITestResult(
            inserter.get_id, stdout, previous_run)
        summary_result = output_result.get_summary()
    result = testtools.CopyStreamResult([inserter, output_result])
    result.startTestRun()
    try:
        case.run(result)
    finally:
        result.stopTestRun()
    if pretty_out and not subunit_out:
        start_times = []
        stop_times = []
        for worker in subunit_trace.RESULTS:
            start_times += [
                x['timestamps'][0] for x in subunit_trace.RESULTS[worker]]
            stop_times += [
                x['timestamps'][1] for x in subunit_trace.RESULTS[worker]]
        start_time = min(start_times)
        stop_time = max(stop_times)
        elapsed_time = stop_time - start_time
        subunit_trace.print_fails(stdout)
        subunit_trace.print_summary(stdout, elapsed_time)
    if not results.wasSuccessful(summary_result):
        return 1
    else:
        return 0



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

import six

if not os.path.isfile('.testr.conf'):
    sys.exit("Testr config file not found")

with open('.testr.conf', 'r') as testr_conf_file:
    config = six.moves.configparser.ConfigParser()
    config.readfp(testr_conf_file)

    test_command = config.get('DEFAULT', 'test_command')
    group_regex = None
    if config.has_option('DEFAULT', 'group_regex'):
        group_regex = config.get('DEFAULT', 'group_regex')

top_dir = None
test_dir = None
for line in test_command.split('\n'):
    if 'subunit.run discover' in line:
        command_parts = line.split(' ')
        top_dir_present = '-t' in line
        for idx, val in enumerate(command_parts):
            if top_dir_present:
                if val == '-t':
                    top_dir = command_parts[idx + 1]
                    test_dir = command_parts[idx + 2]
            else:
                if val == 'discover':
                    test_dir = command_parts[idx + 1]

with open('.stestr.conf', 'w') as stestr_conf_file:
    stestr_conf_file.write('[DEFAULT]\n')
    stestr_conf_file.write('test_path=%s\n' % test_dir)
    if top_dir:
        stestr_conf_file.write('top_dir=%s\n' % top_dir)
    if group_regex:
        stestr_conf_file.write('group_regex=%s\n' % group_regex)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initialise a new repository."""

import sys

from cliff import command

from stestr.repository import util


class Init(command.Command):
    """Create a new repository."""

    def take_action(self, parsed_args):
        init(self.app_args.repo_type, self.app_args.repo_url)


def init(repo_type='file', repo_url=None, stdout=sys.stdout):
    """Initialize a new repository

    This function will create initialize a new repostiory if one does not
    exist. If one exists the command will fail.

    Note this function depends on the cwd for the repository if `repo_type` is
    set to file and `repo_url` is not specified it will use the repository
    located at CWD/.stestr

    :param str repo_type: This is the type of repository to use. Valid choices
        are 'file' and 'sql'.
    :param str repo_url: The url of the repository to use.

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """
    try:
        util.get_repo_initialise(repo_type, repo_url)
    except OSError as e:
        if e.errno != 17:
            raise
        repo_path = repo_url or './stestr'
        stdout.write('The specified repository directory %s already exists. '
                     'Please check if the repository already exists or '
                     'select a different path\n' % repo_path)
        return 1



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Show the current failures in the repository."""

import sys

from cliff import command
import testtools

from stestr import output
from stestr.repository import util
from stestr import results
from stestr import user_config


class Failing(command.Command):
    """Show the current failures known by the repository.

    Without --subunit, the process exit code will be non-zero if the
    previous test run was not successful and test failures are shown. But,
    with --subunit, the process exit code is non-zero only if the subunit
    stream could not be generated successfully from any failures. The test
    results and run status are included in the subunit stream emitted for
    the failed tests, so the stream should be used for interpretting the
    failing tests. If no subunit stream is emitted with --subunit and a
    zero exit code then there were no failures in the most recent run in
    the repository.
    """

    def get_parser(self, prog_name):
        parser = super(Failing, self).get_parser(prog_name)
        parser.add_argument(
            "--subunit", action="store_true",
            default=False, help="Show output as a subunit stream.")
        parser.add_argument(
            "--list", action="store_true",
            default=False, help="Show only a list of failing tests.")
        return parser

    def take_action(self, parsed_args):
        user_conf = user_config.get_user_config(self.app_args.user_config)
        args = parsed_args
        if getattr(user_conf, 'failing', False):
            list_opt = args.list or user_conf.failing.get('list', False)
        else:
            list_opt = args.list
        return failing(repo_type=self.app_args.repo_type,
                       repo_url=self.app_args.repo_url,
                       list_tests=list_opt, subunit=args.subunit)


def _show_subunit(run):
    stream = run.get_subunit_stream()
    if getattr(sys.stdout, 'buffer', False):
        sys.stdout.buffer.write(stream.read())
    else:
        sys.stdout.write(stream.read())
    return 0


def _make_result(repo, list_tests=False, stdout=sys.stdout):
    if list_tests:
        list_result = testtools.StreamSummary()
        return list_result, list_result
    else:
        def _get_id():
            return repo.get_latest_run().get_id()

        output_result = results.CLITestResult(_get_id,
                                              stdout, None)
        summary_result = output_result.get_summary()
        return output_result, summary_result


def failing(repo_type='file', repo_url=None, list_tests=False, subunit=False,
            stdout=sys.stdout):
    """Print the failing tests from the most recent run in the repository

    This function will print to STDOUT whether there are any tests that failed
    in the last run. It optionally will print the test_ids for the failing
    tests if ``list_tests`` is true. If ``subunit`` is true a subunit stream
    with just the failed tests will be printed to STDOUT.

    Note this function depends on the cwd for the repository if `repo_type` is
    set to file and `repo_url` is not specified it will use the repository
    located at CWD/.stestr

    :param str repo_type: This is the type of repository to use. Valid choices
        are 'file' and 'sql'.
    :param str repo_url: The url of the repository to use.
    :param bool list_test: Show only a list of failing tests.
    :param bool subunit: Show output as a subunit stream.
    :param file stdout: The output file to write all output to. By default
        this is sys.stdout

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """
    if repo_type not in ['file', 'sql']:
        stdout.write('Repository type %s is not a type' % repo_type)
        return 1

    repo = util.get_repo_open(repo_type, repo_url)
    run = repo.get_failing()
    if subunit:
        return _show_subunit(run)
    case = run.get_test()
    failed = False
    result, summary = _make_result(repo, list_tests=list_tests)
    result.startTestRun()
    try:
        case.run(result)
    finally:
        result.stopTestRun()
    failed = not results.wasSuccessful(summary)
    if failed:
        result = 1
    else:
        result = 0
    if list_tests:
        failing_tests = [
            test for test, _ in summary.errors + summary.failures]
        output.output_tests(failing_tests, output=stdout)
    return result



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six
import subunit
import testtools

from stestr import output


def wasSuccessful(summary):
    return not (summary.errors or summary.failures or
                summary.unexpectedSuccesses)


class SummarizingResult(testtools.StreamSummary):

    def __init__(self):
        super(SummarizingResult, self).__init__()

    def startTestRun(self):
        super(SummarizingResult, self).startTestRun()
        self._first_time = None
        self._last_time = None

    def status(self, *args, **kwargs):
        if kwargs.get('timestamp') is not None:
            timestamp = kwargs['timestamp']
            if self._last_time is None:
                self._first_time = timestamp
                self._last_time = timestamp
            if timestamp < self._first_time:
                self._first_time = timestamp
            if timestamp > self._last_time:
                self._last_time = timestamp
        super(SummarizingResult, self).status(*args, **kwargs)

    def get_num_failures(self):
        return len(self.failures) + len(self.errors)

    def get_time_taken(self):
        if None in (self._last_time, self._first_time):
            return None
        return (self._last_time - self._first_time).total_seconds()


class CatFiles(testtools.StreamResult):
    """Cat file attachments received to a stream."""

    def __init__(self, byte_stream):
        self.stream = subunit.make_stream_binary(byte_stream)
        self.last_file = None

    def status(self, test_id=None, test_status=None, test_tags=None,
               runnable=True, file_name=None, file_bytes=None, eof=False,
               mime_type=None, route_code=None, timestamp=None):
        if file_name is None:
            return
        if self.last_file != file_name:
            self.stream.write(("--- %s ---\n" % file_name).encode('utf8'))
            self.last_file = file_name
        self.stream.write(file_bytes)
        self.stream.flush()


class CLITestResult(testtools.StreamResult):
    """A TestResult for the CLI."""

    def __init__(self, get_id, stream, previous_run=None):
        """Construct a CLITestResult writing to stream.

        :param get_id: A nullary callable that returns the id of the test run.
                       This expects a callable instead of the actual value
                       because in some repository backends the run_id is only
                       generated after stopTestRun() is called.
        :param stream: The stream to use for result
        :param previous_run: The CLITestResult for the previous run
        """
        super(CLITestResult, self).__init__()
        self._previous_run = previous_run
        self._summary = SummarizingResult()
        self.stream = testtools.compat.unicode_output_stream(stream)
        self.sep1 = testtools.compat._u('=' * 70 + '\n')
        self.sep2 = testtools.compat._u('-' * 70 + '\n')
        self.filterable_states = set(['success', 'uxsuccess', 'xfail', 'skip'])
        self.get_id = get_id

    def _format_error(self, label, test, error_text, test_tags=None):
        test_tags = test_tags or ()
        tags = ' '.join(test_tags)
        if tags:
            tags = six.text_type(('tags: %s\n' % tags))
        return six.text_type(''.join([
            self.sep1,
            six.text_type('%s: %s\n' % (label, test.id())),
            tags,
            self.sep2,
            error_text,
            ]))

    def status(self, **kwargs):
        super(CLITestResult, self).status(**kwargs)
        self._summary.status(**kwargs)
        test_status = kwargs.get('test_status')
        test_tags = kwargs.get('test_tags')
        if test_status == 'fail':
            self.stream.write(
                self._format_error(six.text_type('FAIL'),
                                   *(self._summary.errors[-1]),
                                   test_tags=test_tags))
        if test_status not in self.filterable_states:
            return

    def _get_previous_summary(self):
        if self._previous_run is None:
            return None
        previous_summary = SummarizingResult()
        previous_summary.startTestRun()
        test = self._previous_run.get_test()
        test.run(previous_summary)
        previous_summary.stopTestRun()
        return previous_summary

    def _output_summary(self, run_id):
        """Output a test run.

        :param run_id: The run id.
        """
        time = self._summary.get_time_taken()
        time_delta = None
        num_tests_run_delta = None
        num_failures_delta = None
        values = [('id', run_id, None)]
        failures = self._summary.get_num_failures()
        previous_summary = self._get_previous_summary()
        if failures:
            if previous_summary:
                num_failures_delta = failures - \
                    previous_summary.get_num_failures()
            values.append(('failures', failures, num_failures_delta))
        if previous_summary:
            num_tests_run_delta = self._summary.testsRun - \
                previous_summary.testsRun
            if time:
                previous_time_taken = previous_summary.get_time_taken()
                if previous_time_taken:
                    time_delta = time - previous_time_taken
        skips = len(self._summary.skipped)
        if skips:
            values.append(('skips', skips, None))
        output.output_summary(
            not bool(failures), self._summary.testsRun, num_tests_run_delta,
            time, time_delta, values, output=self.stream)

    def startTestRun(self):
        super(CLITestResult, self).startTestRun()
        self._summary.startTestRun()

    def stopTestRun(self):
        super(CLITestResult, self).stopTestRun()
        run_id = self.get_id()
        self._summary.stopTestRun()
        self._output_summary(run_id)

    def get_summary(self):
        return self._summary



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import re

import mock
from subunit import iso8601

from stestr.repository import memory
from stestr import scheduler
from stestr.tests import base


class TestScheduler(base.TestCase):

    def _add_timed_test(self, id, duration, result):
        start = datetime.datetime.now()
        start = start.replace(tzinfo=iso8601.UTC)
        result.status(test_id=id, test_status='inprogress',
                      timestamp=start)
        timestamp = start + datetime.timedelta(seconds=duration)
        result.status(test_id=id, test_status='success',
                      timestamp=timestamp)

    def test_partition_tests(self):
        repo = memory.RepositoryFactory().initialise('memory:')
        result = repo.get_inserter()
        result.startTestRun()
        self._add_timed_test("slow", 3, result)
        self._add_timed_test("fast1", 1, result)
        self._add_timed_test("fast2", 1, result)
        result.stopTestRun()
        test_ids = frozenset(['slow', 'fast1', 'fast2', 'unknown1',
                              'unknown2', 'unknown3', 'unknown4'])
        partitions = scheduler.partition_tests(test_ids, 2, repo, None)
        self.assertTrue('slow' in partitions[0])
        self.assertFalse('fast1' in partitions[0])
        self.assertFalse('fast2' in partitions[0])
        self.assertFalse('slow' in partitions[1])
        self.assertTrue('fast1' in partitions[1])
        self.assertTrue('fast2' in partitions[1])
        self.assertEqual(3, len(partitions[0]))
        self.assertEqual(4, len(partitions[1]))

    def test_random_partitions(self):
        repo = memory.RepositoryFactory().initialise('memory:')
        test_ids = frozenset(['a_test', 'b_test', 'c_test', 'd_test'])
        random_parts = scheduler.partition_tests(test_ids, 2, repo, None,
                                                 randomize=True)
        # NOTE(masayukig): We can't test this randomness. So just checking
        # what we should get here.
        self.assertEqual(2, len(random_parts))
        self.assertTrue(isinstance(random_parts, list))
        self.assertTrue(isinstance(random_parts[0], list))
        self.assertTrue(isinstance(random_parts[1], list))
        flatten_random_parts = []
        for i, j in random_parts:
            flatten_random_parts.append(i)
            flatten_random_parts.append(j)
        for i in test_ids:
            self.assertIn(i, flatten_random_parts)

    def test_partition_tests_with_zero_duration(self):
        repo = memory.RepositoryFactory().initialise('memory:')
        result = repo.get_inserter()
        result.startTestRun()
        self._add_timed_test("zero1", 0, result)
        self._add_timed_test("zero2", 0, result)
        result.stopTestRun()
        # Partitioning by two should generate two one-entry partitions.
        test_ids = frozenset(['zero1', 'zero2'])
        partitions = scheduler.partition_tests(test_ids, 2, repo, None)
        self.assertEqual(1, len(partitions[0]))
        self.assertEqual(1, len(partitions[1]))

    def test_partition_tests_with_grouping(self):
        repo = memory.RepositoryFactory().initialise('memory:')
        result = repo.get_inserter()
        result.startTestRun()
        self._add_timed_test("TestCase1.slow", 3, result)
        self._add_timed_test("TestCase2.fast1", 1, result)
        self._add_timed_test("TestCase2.fast2", 1, result)
        result.stopTestRun()
        test_ids = frozenset(['TestCase1.slow', 'TestCase1.fast',
                              'TestCase1.fast2', 'TestCase2.fast1',
                              'TestCase3.test1', 'TestCase3.test2',
                              'TestCase2.fast2', 'TestCase4.test',
                              'testdir.testfile.TestCase5.test'])

        def group_id(test_id, regex=re.compile('TestCase[0-5]')):
            match = regex.match(test_id)
            if match:
                return match.group(0)

        partitions = scheduler.partition_tests(test_ids, 2, repo, group_id)
        # Timed groups are deterministic:
        self.assertTrue('TestCase2.fast1' in partitions[0])
        self.assertTrue('TestCase2.fast2' in partitions[0])
        self.assertTrue('TestCase1.slow' in partitions[1])
        self.assertTrue('TestCase1.fast' in partitions[1])
        self.assertTrue('TestCase1.fast2' in partitions[1])
        # Untimed groups just need to be in the same partition:
        if 'TestCase3.test1' in partitions[0]:
            self.assertTrue('TestCase3.test2' in partitions[0])
        if 'TestCase4.test' not in partitions[0]:
            self.assertTrue('TestCase4.test' in partitions[1])
        if 'testdir.testfile.TestCase5.test' not in partitions[0]:
            self.assertTrue('testdir.testfile.TestCase5.test' in partitions[1])

    @mock.patch('six.moves.builtins.open', mock.mock_open(), create=True)
    def test_generate_worker_partitions(self):
        test_ids = ['test_a', 'test_b', 'your_test']
        fake_worker_yaml = [
            {'worker': ['test_']},
            {'worker': ['test']},
        ]
        with mock.patch('yaml.load', return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, 'fakepath')
        expected_grouping = [
            ['test_a', 'test_b'],
            ['test_a', 'test_b', 'your_test'],
        ]
        self.assertEqual(expected_grouping, groups)

    @mock.patch('six.moves.builtins.open', mock.mock_open(), create=True)
    def test_generate_worker_partitions_group_without_list(self):
        test_ids = ['test_a', 'test_b', 'your_test']
        fake_worker_yaml = [
            {'worker': ['test_']},
            {'worker': 'test'},
        ]
        with mock.patch('yaml.load', return_value=fake_worker_yaml):
            self.assertRaises(TypeError, scheduler.generate_worker_partitions,
                              test_ids, 'fakepath')

    @mock.patch('six.moves.builtins.open', mock.mock_open(), create=True)
    def test_generate_worker_partitions_no_worker_tag(self):
        test_ids = ['test_a', 'test_b', 'your_test']
        fake_worker_yaml = [
            {'worker-foo': ['test_']},
            {'worker': ['test']},
        ]
        with mock.patch('yaml.load', return_value=fake_worker_yaml):
            self.assertRaises(TypeError, scheduler.generate_worker_partitions,
                              test_ids, 'fakepath')

    @mock.patch('six.moves.builtins.open', mock.mock_open(), create=True)
    def test_generate_worker_partitions_group_without_match(self):
        test_ids = ['test_a', 'test_b', 'your_test']
        fake_worker_yaml = [
            {'worker': ['test_']},
            {'worker': ['test']},
            {'worker': ['foo']}
        ]
        with mock.patch('yaml.load', return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, 'fakepath')
        expected_grouping = [
            ['test_a', 'test_b'],
            ['test_a', 'test_b', 'your_test'],
        ]
        self.assertEqual(expected_grouping, groups)

    @mock.patch('six.moves.builtins.open', mock.mock_open(), create=True)
    def test_generate_worker_partitions_with_count(self):
        test_ids = ['test_a', 'test_b', 'your_test', 'a_thing1', 'a_thing2']
        fake_worker_yaml = [
            {'worker': ['test_']},
            {'worker': ['test']},
            {'worker': ['a_thing'], 'concurrency': 2},
        ]
        with mock.patch('yaml.load', return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, 'fakepath')
        expected_grouping = [
            ['test_a', 'test_b'],
            ['test_a', 'test_b', 'your_test'],
            ['a_thing1'],
            ['a_thing2'],
        ]
        for worker in expected_grouping:
            self.assertIn(worker, groups)

    @mock.patch('six.moves.builtins.open', mock.mock_open(), create=True)
    def test_generate_worker_partitions_with_count_1(self):
        test_ids = ['test_a', 'test_b', 'your_test']
        fake_worker_yaml = [
            {'worker': ['test_']},
            {'worker': ['test'], 'count': 1},
        ]
        with mock.patch('yaml.load', return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, 'fakepath')
        expected_grouping = [
            ['test_a', 'test_b'],
            ['test_a', 'test_b', 'your_test'],
        ]
        self.assertEqual(expected_grouping, groups)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Tests for the file repository implementation."""

import os.path
import shutil
import tempfile

import fixtures
import testtools
from testtools import matchers

from stestr.repository import file
from stestr.tests import base


class FileRepositoryFixture(fixtures.Fixture):

    def __init__(self, path=None):
        super(FileRepositoryFixture, self).__init__()
        self.path = path

    def setUp(self):
        super(FileRepositoryFixture, self).setUp()
        if self.path and os.path.isdir(self.path):
            self.tempdir = self.path
        else:
            self.tempdir = tempfile.mkdtemp()
            self.addCleanup(shutil.rmtree, self.tempdir)
        self.repo = file.RepositoryFactory().initialise(self.tempdir)


class HomeDirTempDir(fixtures.Fixture):
    """Creates a temporary directory in ~."""

    def setUp(self):
        super(HomeDirTempDir, self).setUp()
        home_dir = os.path.expanduser('~')
        self.temp_dir = tempfile.mkdtemp(dir=home_dir)
        self.addCleanup(shutil.rmtree, self.temp_dir)
        self.short_path = os.path.join('~', os.path.basename(self.temp_dir))


class TestFileRepository(base.TestCase):

    def setUp(self):
        super(TestFileRepository, self).setUp()
        self.tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tempdir)

    def test_initialise(self):
        self.useFixture(FileRepositoryFixture(path=self.tempdir))
        base = os.path.join(self.tempdir, '.stestr')
        self.assertTrue(os.path.isdir(base))
        self.assertTrue(os.path.isfile(os.path.join(base, 'format')))
        with open(os.path.join(base, 'format'), 'rt') as stream:
            contents = stream.read()
        self.assertEqual("1\n", contents)
        with open(os.path.join(base, 'next-stream'), 'rt') as stream:
            contents = stream.read()
        self.assertEqual("0\n", contents)

    # Skip if windows since ~ in a path doesn't work there
    @testtools.skipIf(os.name == 'nt', "Windows doesn't support '~' expand")
    def test_initialise_expands_user_directory(self):
        short_path = self.useFixture(HomeDirTempDir()).short_path
        repo = file.RepositoryFactory().initialise(short_path)
        self.assertTrue(os.path.exists(repo.base))

    def test_inserter_output_path(self):
        repo = self.useFixture(FileRepositoryFixture()).repo
        inserter = repo.get_inserter()
        inserter.startTestRun()
        inserter.stopTestRun()
        self.assertTrue(os.path.exists(os.path.join(repo.base, '0')))

    def test_inserting_creates_id(self):
        # When inserting a stream, an id is returned from stopTestRun.
        repo = self.useFixture(FileRepositoryFixture()).repo
        result = repo.get_inserter()
        result.startTestRun()
        result.stopTestRun()
        self.assertEqual(0, result.get_id())

    # Skip if windows since ~ in a path doesn't work there
    @testtools.skipIf(os.name == 'nt', "Windows doesn't support '~' expand")
    def test_open_expands_user_directory(self):
        short_path = self.useFixture(HomeDirTempDir()).short_path
        repo1 = file.RepositoryFactory().initialise(short_path)
        repo2 = file.RepositoryFactory().open(short_path)
        self.assertEqual(repo1.base, repo2.base)

    def test_next_stream_corruption_error(self):
        repo = self.useFixture(FileRepositoryFixture()).repo
        open(os.path.join(repo.base, 'next-stream'), 'wb').close()
        self.assertThat(repo.count, matchers.Raises(
            matchers.MatchesException(
                ValueError("Corrupt next-stream file: ''"))))

    # Skip if windows since chmod doesn't work there
    @testtools.skipIf(os.name == 'nt', "Windows doesn't support chmod")
    def test_get_test_run_unexpected_ioerror_errno(self):
        repo = self.useFixture(FileRepositoryFixture()).repo
        inserter = repo.get_inserter()
        inserter.startTestRun()
        inserter.stopTestRun()
        self.assertTrue(os.path.isfile(os.path.join(repo.base, '0')))
        os.chmod(os.path.join(repo.base, '0'), 0000)
        self.assertRaises(IOError, repo.get_test_run, '0')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Persistent storage of test results."""
from __future__ import print_function

import datetime
import io
import os.path
import re
import subprocess
import sys

import six
import sqlalchemy
from sqlalchemy import orm
import subunit.v2
from subunit2sql.db import api as db_api
from subunit2sql import read_subunit
from subunit2sql import shell
from subunit2sql import write_subunit
import testtools

from stestr.repository import abstract as repository
from stestr import utils


def atomicish_rename(source, target):
    if os.name != "posix" and os.path.exists(target):
        os.remove(target)
    os.rename(source, target)


class RepositoryFactory(repository.AbstractRepositoryFactory):

    def initialise(klass, url):
        """Create a repository at url/path."""
        print("WARNING: The SQL repository type is still experimental. You "
              "might encounter issues while using it.", file=sys.stderr)
        result = Repository(url)
        # TODO(mtreinish): Figure out the python api to run the migrations for
        # setting up the schema.
        proc = subprocess.Popen(['subunit2sql-db-manage',
                                 '--database-connection', url, 'upgrade',
                                 'head'],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        sys.stdout.write(six.text_type(out))
        sys.stderr.write(six.text_type(err))

        return result

    def open(self, url):
        repo = Repository(url)
        # To test the repository's existence call get_ids_for_all_tests()
        # if it raises an OperationalError that means the DB doesn't exist or
        # it couldn't connect, either way the repository was not found.
        try:
            session = repo.session_factory()
            db_api.get_ids_for_all_tests(session=session)
            session.close()
        except sqlalchemy.exc.OperationalError:
            raise repository.RepositoryNotFound(url)
        return repo


class Repository(repository.AbstractRepository):
    """subunit2sql based storage of test results.

    This repository stores each stream in a subunit2sql DB. Refer to the
    subunit2sql documentation for
    """

    def __init__(self, url):
        """Create a subunit2sql-based repository object for the repo at 'url'.

        :param base: The path to the repository.
        """
        self.base = url
        self.engine = sqlalchemy.create_engine(url)
        self.session_factory = orm.sessionmaker(bind=self.engine)

    # TODO(mtreinish): We need to add a subunit2sql api to get the count
    def count(self):
        super(Repository, self).count()

    def _get_latest_run(self):
        session = self.session_factory()
        latest_run = db_api.get_latest_run(session)
        session.close()
        if not latest_run:
            raise KeyError("No tests in repository")
        return latest_run

    def latest_id(self):
        return self._get_latest_run().uuid

    def get_failing(self):
        latest_run = self._get_latest_run()
        session = self.session_factory()
        failed_test_runs = db_api.get_test_runs_by_status_for_run_ids(
            'fail', [latest_run.id], session=session)
        session.close()
        return _Subunit2SqlRun(self.base, None, test_runs=failed_test_runs)

    def get_test_run(self, run_id):
        return _Subunit2SqlRun(self.base, run_id)

    def _get_inserter(self, partial, run_id=None):
        return _SqlInserter(self, partial, run_id)

    def _get_test_times(self, test_ids):
        result = {}
        # TODO(mtreinish): after subunit2sql adds a bulk query for getting
        # multiple tests by test_id at once remove the for loop
        session = self.session_factory()
        for test_id in test_ids:
            stripped_test_id = utils.cleanup_test_name(test_id)
            test = db_api.get_test_by_test_id(stripped_test_id,
                                              session=session)
            if test:
                # NOTE(mtreinish): We need to make sure the test_id with attrs
                # is used in the output dict, otherwise the scheduler won't
                # see it
                result[test_id] = test.run_time
        session.close()
        return result


class _Subunit2SqlRun(repository.AbstractTestRun):
    """A test run that was inserted into the repository."""

    def __init__(self, url, run_id, test_runs=None):
        engine = sqlalchemy.create_engine(url)
        self.session_factory = orm.sessionmaker(bind=engine)
        self._run_id = run_id
        self._test_runs = test_runs

    def get_id(self):
        return self._run_id

    def get_subunit_stream(self):
        stream = io.BytesIO()
        if self._run_id:
            session = self.session_factory()
            test_runs = db_api.get_tests_run_dicts_from_run_id(self._run_id,
                                                               session)
            session.close()
        else:
            test_runs = self._test_runs
        output = subunit.v2.StreamResultToBytes(stream)
        output.startTestRun()
        for test_id in test_runs:
            test = test_runs[test_id]
            # NOTE(mtreinish): test_run_metadata is not guaranteed to be
            # present for the test run
            metadata = test.get('metadata', None)
            write_subunit.write_test(output, test['start_time'],
                                     test['stop_time'], test['status'],
                                     test_id, metadata)
        output.stopTestRun()
        stream.seek(0)
        return stream

    def get_test(self):
        stream = self.get_subunit_stream()
        case = subunit.ByteStreamToStreamResult(stream)
        return case


class _SqlInserter(repository.AbstractTestRun):
    """Insert test results into a sql repository."""

    def __init__(self, repository, partial=False, run_id=None):
        self._repository = repository
        self.partial = partial
        self._subunit = None
        self._run_id = run_id
        # Create a new session factory
        self.engine = sqlalchemy.create_engine(self._repository.base)
        self.session_factory = orm.sessionmaker(bind=self.engine,
                                                autocommit=True)

    def startTestRun(self):
        self._subunit = io.BytesIO()
        self.subunit_stream = subunit.v2.StreamResultToBytes(self._subunit)
        self.hook = testtools.CopyStreamResult([
            testtools.StreamToDict(self._handle_test),
            self.subunit_stream])
        self.hook.startTestRun()
        self.start_time = datetime.datetime.utcnow()
        session = self.session_factory()
        if not self._run_id:
            self.run = db_api.create_run(session=session)
            self._run_id = self.run.uuid
        else:
            int_id = db_api.get_run_id_from_uuid(self._run_id, session=session)
            self.run = db_api.get_run_by_id(int_id, session=session)
        session.close()
        self.totals = {}

    def _update_test(self, test_dict, session, start_time, stop_time):
        test_id = utils.cleanup_test_name(test_dict['id'])
        db_test = db_api.get_test_by_test_id(test_id, session)
        if not db_test:
            if test_dict['status'] == 'success':
                success = 1
                fails = 0
            elif test_dict['status'] == 'fail':
                fails = 1
                success = 0
            else:
                fails = 0
                success = 0
            run_time = read_subunit.get_duration(start_time, stop_time)
            db_test = db_api.create_test(test_id, (success + fails), success,
                                         fails, run_time,
                                         session)
        else:
            test_dict['start_time'] = start_time
            test_dict['end_time'] = stop_time
            test_values = shell.increment_counts(db_test, test_dict)
            # If skipped nothing to update
            if test_values:
                db_api.update_test(test_values, db_test.id, session)
        return db_test

    def _get_attrs(self, test_id):
        attr_regex = re.compile('\[(.*)\]')
        matches = attr_regex.search(test_id)
        attrs = None
        if matches:
            attrs = matches.group(1)
        return attrs

    def _handle_test(self, test_dict):
        start, end = test_dict.pop('timestamps')
        if test_dict['status'] == 'exists' or None in (start, end):
            return
        elif test_dict['id'] == 'process-returncode':
            return
        session = self.session_factory()
        try:
            # Update the run counts
            if test_dict['status'] not in self.totals:
                self.totals[test_dict['status']] = 1
            else:
                self.totals[test_dict['status']] += 1
            values = {}
            if test_dict['status'] in ('success', 'xfail'):
                values['passes'] = self.totals['success']
            elif test_dict['status'] in ('fail', 'uxsuccess'):
                values['fails'] = self.totals['fail']
            elif test_dict['status'] == 'skip':
                values['skips'] = self.totals['skip']
            db_api.update_run(values, self.run.id, session=session)
            # Update the test totals
            db_test = self._update_test(test_dict, session, start,
                                        end)
            # Add the test run
            test_run = db_api.create_test_run(db_test.id, self.run.id,
                                              test_dict['status'],
                                              start, end,
                                              session)
            metadata = {}
            attrs = self._get_attrs(test_dict['id'])
            if attrs:
                metadata['attrs'] = attrs
            if test_dict.get('tags', None):
                metadata['tags'] = ",".join(test_dict['tags'])
            if metadata:
                db_api.add_test_run_metadata(
                    metadata, test_run.id, session)
            # TODO(mtreinish): Add attachments support to the DB.
            session.close()
        except Exception:
            session.rollback()
            raise

    def stopTestRun(self):
        self.hook.stopTestRun()
        stop_time = datetime.datetime.utcnow()
        self._subunit.seek(0)
        values = {}
        values['run_time'] = read_subunit.get_duration(self.start_time,
                                                       stop_time)
        session = self.session_factory()
        db_api.update_run(values, self.run.id, session=session)
        session.close()

    def status(self, *args, **kwargs):
        self.hook.status(*args, **kwargs)

    def get_id(self):
        return self._run_id



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Tests for the sql repository implementation."""

import os
import os.path
import tempfile
import uuid

import fixtures

from stestr.repository import sql
from stestr.tests import base


class SqlRepositoryFixture(fixtures.Fixture):

    def __init__(self, url=None):
        super(SqlRepositoryFixture, self).__init__()
        self.url = url

    def setUp(self):
        super(SqlRepositoryFixture, self).setUp()
        self.repo = sql.RepositoryFactory().initialise(self.url)


class TestSqlRepository(base.TestCase):

    def setUp(self):
        super(TestSqlRepository, self).setUp()
        # NOTE(mtreinish): Windows likes to fail if the file is already open
        # when we access it later, so lets explicitly close it before we move
        # forward
        _close_me, self.tempfile = tempfile.mkstemp(suffix='.sqlite')
        os.close(_close_me)
        self.addCleanup(os.remove, self.tempfile)
        self.url = 'sqlite:///' + self.tempfile

    def test_initialise(self):
        self.useFixture(SqlRepositoryFixture(url=self.url))

    def test_get_failing(self):
        repo = self.useFixture(SqlRepositoryFixture(url=self.url)).repo
        # NOTE: "No tests in repository"
        self.assertRaises(KeyError, repo.get_failing)
        inserter = repo.get_inserter()
        inserter.startTestRun()
        inserter.stopTestRun()
        self.assertIsInstance(repo.get_failing(), sql._Subunit2SqlRun)

    def test_inserter_output_path(self):
        repo = self.useFixture(SqlRepositoryFixture(url=self.url)).repo
        inserter = repo.get_inserter()
        inserter.startTestRun()
        inserter.stopTestRun()
        run_id = inserter.get_id()
        run_uuid = uuid.UUID(run_id)

        self.assertEqual(uuid.UUID(repo.latest_id()), run_uuid)

    def test_run_get_subunit_stream(self):
        repo = self.useFixture(SqlRepositoryFixture(url=self.url)).repo
        inserter = repo.get_inserter()
        inserter.startTestRun()
        inserter.stopTestRun()
        run_id = inserter.get_id()
        run = repo.get_test_run(run_id)

        stream = run.get_subunit_stream()
        self.assertIsNotNone(stream)
        self.assertTrue(stream.readable())
        self.assertEqual([], stream.readlines())



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import io
import os
import re
import shutil
import subprocess
import tempfile

import fixtures
import six
from six import StringIO
import subunit as subunit_lib
import testtools

from stestr.commands import list as list_cmd
from stestr.commands import run
from stestr.tests import base


class TestReturnCodes(base.TestCase):
    def setUp(self):
        super(TestReturnCodes, self).setUp()
        # Setup test dirs
        self.directory = tempfile.mkdtemp(prefix='stestr-unit')
        self.addCleanup(shutil.rmtree, self.directory, ignore_errors=True)
        self.test_dir = os.path.join(self.directory, 'tests')
        os.mkdir(self.test_dir)
        # Setup Test files
        self.testr_conf_file = os.path.join(self.directory, '.stestr.conf')
        self.setup_cfg_file = os.path.join(self.directory, 'setup.cfg')
        self.passing_file = os.path.join(self.test_dir, 'test_passing.py')
        self.failing_file = os.path.join(self.test_dir, 'test_failing.py')
        self.init_file = os.path.join(self.test_dir, '__init__.py')
        self.setup_py = os.path.join(self.directory, 'setup.py')
        self.user_config = os.path.join(self.directory, 'stestr.yaml')
        shutil.copy('stestr/tests/files/testr-conf', self.testr_conf_file)
        shutil.copy('stestr/tests/files/passing-tests', self.passing_file)
        shutil.copy('stestr/tests/files/failing-tests', self.failing_file)
        shutil.copy('setup.py', self.setup_py)
        shutil.copy('stestr/tests/files/setup.cfg', self.setup_cfg_file)
        shutil.copy('stestr/tests/files/__init__.py', self.init_file)
        shutil.copy('stestr/tests/files/stestr.yaml', self.user_config)

        self.stdout = StringIO()
        self.stderr = StringIO()
        # Change directory, run wrapper and check result
        self.addCleanup(os.chdir, os.path.abspath(os.curdir))
        os.chdir(self.directory)
        subprocess.call('stestr init', shell=True)

    def _check_subunit(self, output_stream):
        stream = subunit_lib.ByteStreamToStreamResult(output_stream)
        starts = testtools.StreamResult()
        summary = testtools.StreamSummary()
        tests = []

        def _add_dict(test):
            tests.append(test)

        outcomes = testtools.StreamToDict(functools.partial(_add_dict))
        result = testtools.CopyStreamResult([starts, outcomes, summary])
        result.startTestRun()
        try:
            stream.run(result)
        finally:
            result.stopTestRun()
        self.assertThat(len(tests), testtools.matchers.GreaterThan(0))

    def assertRunExit(self, cmd, expected, subunit=False, stdin=None):
        if stdin:
            p = subprocess.Popen(
                "%s" % cmd, shell=True, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate(stdin)
        else:
            p = subprocess.Popen(
                "%s" % cmd, shell=True,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()

        if not subunit:
            self.assertEqual(
                p.returncode, expected,
                "Stdout: %s; Stderr: %s" % (out, err))
            return (out, err)
        else:
            self.assertEqual(p.returncode, expected,
                             "Expected return code: %s doesn't match actual "
                             "return code of: %s" % (expected, p.returncode))
            output_stream = io.BytesIO(out)
            stream = subunit_lib.ByteStreamToStreamResult(output_stream)
            starts = testtools.StreamResult()
            summary = testtools.StreamSummary()
            tests = []

            def _add_dict(test):
                tests.append(test)

            outcomes = testtools.StreamToDict(functools.partial(_add_dict))
            result = testtools.CopyStreamResult([starts, outcomes, summary])
            result.startTestRun()
            try:
                stream.run(result)
            finally:
                result.stopTestRun()
            self.assertThat(len(tests), testtools.matchers.GreaterThan(0))
            return (out, err)

    def test_parallel_passing(self):
        self.assertRunExit('stestr run passing', 0)

    def test_parallel_passing_bad_regex(self):
        self.assertRunExit('stestr run bad.regex.foobar', 1)

    def test_parallel_fails(self):
        self.assertRunExit('stestr run', 1)

    def test_parallel_passing_xfail(self):
        self.assertRunExit('stestr run xfail', 0)

    def test_parallel_fails_unxsuccess(self):
        self.assertRunExit('stestr run unexpected', 1)

    def test_parallel_blacklist(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as blacklist:
            blacklist.write('fail')
        cmd = 'stestr run --blacklist-file %s' % path
        self.assertRunExit(cmd, 0)

    def test_parallel_whitelist(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as whitelist:
            whitelist.write('passing')
        cmd = 'stestr run --whitelist-file %s' % path
        self.assertRunExit(cmd, 0)

    def test_serial_passing(self):
        self.assertRunExit('stestr run --serial passing', 0)

    def test_serial_fails(self):
        self.assertRunExit('stestr run --serial', 1)

    def test_serial_blacklist(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as blacklist:
            blacklist.write('fail')
        cmd = 'stestr run --serial --blacklist-file %s' % path
        self.assertRunExit(cmd, 0)

    def test_serial_whitelist(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as whitelist:
            whitelist.write('passing')
        cmd = 'stestr run --serial --whitelist-file %s' % path
        self.assertRunExit(cmd, 0)

    def test_serial_subunit_passing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           '--serial passing', 0, subunit=True)

    def test_serial_subunit_failing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           '--serial failing', 0, subunit=True)

    def test_parallel_subunit_passing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           'passing', 0, subunit=True)

    def test_parallel_subunit_failing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           'failing', 0, subunit=True)

    def test_slowest_passing(self):
        self.assertRunExit('stestr run --slowest passing', 0)

    def test_slowest_failing(self):
        self.assertRunExit('stestr run --slowest failing', 1)

    def test_until_failure_fails(self):
        self.assertRunExit('stestr run --until-failure', 1)

    def test_until_failure_with_subunit_fails(self):
        self.assertRunExit('stestr --user-config stestr.yaml run '
                           '--until-failure --subunit', 1, subunit=True)

    def test_with_parallel_class(self):
        # NOTE(masayukig): Ideally, it's better to figure out the
        # difference between with --parallel-class and without
        # --parallel-class. However, it's difficult to make such a
        # test from a command line based test.
        self.assertRunExit('stestr --parallel-class run passing', 0)

    def test_list(self):
        self.assertRunExit('stestr list', 0)

    def _get_cmd_stdout(self, cmd):
        p = subprocess.Popen(cmd, shell=True,
                             stdout=subprocess.PIPE)
        out = p.communicate()
        self.assertEqual(0, p.returncode)
        return out

    def test_combine_results(self):
        self.assertRunExit('stestr run passing', 0)
        stdout = self._get_cmd_stdout(
            'stestr last --no-subunit-trace')
        stdout = six.text_type(stdout[0])
        test_count_split = stdout.split(' ')
        test_count = test_count_split[1]
        test_count = int(test_count)
        id_regex = re.compile('\(id=(.*?)\)')
        test_id = id_regex.search(stdout).group(0)
        self.assertRunExit('stestr run --combine passing', 0)
        combine_stdout = self._get_cmd_stdout(
            'stestr last --no-subunit-trace')[0]
        combine_stdout = six.text_type(combine_stdout)
        combine_test_count_split = combine_stdout.split(' ')
        combine_test_count = combine_test_count_split[1]
        combine_test_count = int(combine_test_count)
        combine_test_id = id_regex.search(combine_stdout).group(0)
        self.assertEqual(test_id, combine_test_id)

        # The test results from running the same tests twice with combine
        # should return a test count 2x as big at the end of the run
        self.assertEqual(test_count * 2, combine_test_count)

    def test_load_from_stdin(self):
        self.assertRunExit('stestr run passing', 0)
        stream = self._get_cmd_stdout(
            'stestr last --subunit')[0]
        self.assertRunExit('stestr load', 0, stdin=stream)

    def test_load_from_stdin_quiet(self):
        out, err = self.assertRunExit('stestr --user-config stestr.yaml -q '
                                      'run passing', 0)
        self.assertEqual(out.decode('utf-8'), '')
        # FIXME(masayukig): We get some warnings when we run a coverage job.
        # So, just ignore 'err' here.
        stream = self._get_cmd_stdout('stestr last --subunit')[0]
        out, err = self.assertRunExit('stestr --user-config stestr.yaml -q '
                                      'load', 0, stdin=stream)
        self.assertEqual(out.decode('utf-8'), '')
        self.assertEqual(err.decode('utf-8'), '')

    def test_no_subunit_trace_force_subunit_trace(self):
        out, err = self.assertRunExit(
            'stestr run --no-subunit-trace --force-subunit-trace passing', 0)
        out = six.text_type(out)
        self.assertNotIn('PASSED (id=0)', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    def test_parallel_passing_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(0, run.run_command(filters=['passing'],
                                            stdout=stdout.stream))

    def test_parallel_passing_bad_regex_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(filters=['bad.regex.foobar'],
                                            stdout=stdout.stream))

    def test_parallel_fails_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(stdout=stdout.stream))

    def test_serial_passing_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(0, run.run_command(filters=['passing'], serial=True,
                                            stdout=stdout.stream))

    def test_serial_fails_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(serial=True, stdout=stdout.stream))

    def test_serial_subunit_passing_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(0, run.run_command(subunit_out=True, serial=True,
                                            filters=['passing'],
                                            stdout=stdout))
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_parallel_subunit_passing_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(0, run.run_command(subunit_out=True,
                                            filters=['passing'],
                                            stdout=stdout))
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_until_failure_fails_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(until_failure=True,
                                            stdout=stdout.stream))

    def test_until_failure_with_subunit_fails_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(1, run.run_command(until_failure=True,
                                            subunit_out=True,
                                            stdout=stdout))
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_list_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(0, list_cmd.list_command(stdout=stdout.stream))

    def test_run_no_discover_pytest_path(self):
        passing_string = 'tests/test_passing.py::FakeTestClass::test_pass_list'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 0)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 1', lines)
        self.assertIn(' - Failed: 0', lines)

    def test_run_no_discover_pytest_path_failing(self):
        passing_string = 'tests/test_failing.py::FakeTestClass::test_pass_list'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 1)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 0', lines)
        self.assertIn(' - Failed: 1', lines)

    def test_run_no_discover_file_path(self):
        passing_string = 'tests/test_passing.py'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 0)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 2', lines)
        self.assertIn(' - Failed: 0', lines)
        self.assertIn(' - Expected Fail: 1', lines)

    def test_run_no_discover_file_path_failing(self):
        passing_string = 'tests/test_failing.py'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 1)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 0', lines)
        self.assertIn(' - Failed: 2', lines)
        self.assertIn(' - Unexpected Success: 1', lines)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import itertools
import multiprocessing
import operator
import random

import yaml

from stestr import selection


def partition_tests(test_ids, concurrency, repository, group_callback,
                    randomize=False):
        """Partition test_ids by concurrency.

        Test durations from the repository are used to get partitions which
        have roughly the same expected runtime. New tests - those with no
        recorded duration - are allocated in round-robin fashion to the
        partitions created using test durations.

        :param list test_ids: The list of test_ids to be partitioned
        :param int concurrency: The concurrency that will be used for running
            the tests. This is the number of partitions that test_ids will be
            split into.
        :param repository: A repository object that
        :param group_callback: A callback function that is used as a scheduler
            hint to group test_ids together and treat them as a single unit for
            scheduling. This function expects a single test_id parameter and it
            will return a group identifier. Tests_ids that have the same group
            identifier will be kept on the same worker.
        :param bool randomize: If true each partition's test order will be
                            randomized

        :return: A list where each element is a distinct subset of test_ids,
            and the union of all the elements is equal to set(test_ids).
        """
        _group_callback = group_callback
        partitions = [list() for i in range(concurrency)]
        timed_partitions = [[0.0, partition] for partition in partitions]
        time_data = {}
        if repository:
            time_data = repository.get_test_times(test_ids)
            timed_tests = time_data['known']
            unknown_tests = time_data['unknown']
        else:
            timed_tests = {}
            unknown_tests = set(test_ids)
        # Group tests: generate group_id -> test_ids.
        group_ids = collections.defaultdict(list)
        if _group_callback is None:
            group_callback = lambda _: None
        else:
            group_callback = _group_callback
        for test_id in test_ids:
            group_id = group_callback(test_id) or test_id
            group_ids[group_id].append(test_id)
        # Time groups: generate three sets of groups:
        # - fully timed dict(group_id -> time),
        # - partially timed dict(group_id -> time) and
        # - unknown (set of group_id)
        # We may in future treat partially timed different for scheduling, but
        # at least today we just schedule them after the fully timed groups.
        timed = {}
        partial = {}
        unknown = []
        for group_id, group_tests in group_ids.items():
            untimed_ids = unknown_tests.intersection(group_tests)
            group_time = sum(
                [timed_tests[test_id]
                    for test_id in untimed_ids.symmetric_difference(
                        group_tests)])
            if not untimed_ids:
                timed[group_id] = group_time
            elif group_time:
                partial[group_id] = group_time
            else:
                unknown.append(group_id)

        # Scheduling is NP complete in general, so we avoid aiming for
        # perfection. A quick approximation that is sufficient for our general
        # needs:
        # sort the groups by time
        # allocate to partitions by putting each group in to the partition with
        # the current (lowest time, shortest length[in tests])
        def consume_queue(groups):
            queue = sorted(
                groups.items(), key=operator.itemgetter(1), reverse=True)
            for group_id, duration in queue:
                timed_partitions[0][0] = timed_partitions[0][0] + duration
                timed_partitions[0][1].extend(group_ids[group_id])
                timed_partitions.sort(key=lambda item: (item[0], len(item[1])))

        consume_queue(timed)
        consume_queue(partial)
        # Assign groups with entirely unknown times in round robin fashion to
        # the partitions.
        for partition, group_id in zip(itertools.cycle(partitions), unknown):
            partition.extend(group_ids[group_id])
        if randomize:
            out_parts = []
            for partition in partitions:
                temp_part = list(partition)
                random.shuffle(temp_part)
                out_parts.append(list(temp_part))
            return out_parts
        else:
            return partitions


def local_concurrency():
    """Get the number of available CPUs on the system.

    :return: An int for the number of cpus. Or None if it couldn't be found
    """

    try:
        return multiprocessing.cpu_count()
    except NotImplementedError:
        # No concurrency logic known.
        return None


def generate_worker_partitions(ids, worker_path, repository=None,
                               group_callback=None, randomize=False):
    """Parse a worker yaml file and generate test groups

    :param list ids: A list of test ids too be partitioned
    :param path worker_path: The path to a worker file
    :param repository: A repository object that will be used for looking up
        timing data. This is optional, and also will only be used for
        scheduling if there is a count field on a worker.
    :param group_callback: A callback function that is used as a scheduler
        hint to group test_ids together and treat them as a single unit for
        scheduling. This function expects a single test_id parameter and it
        will return a group identifier. Tests_ids that have the same group
        identifier will be kept on the same worker. This is optional and
        also will only be used for scheduling if there is a count field on a
        worker.
    :param bool randomize: If true each partition's test order will be
        randomized. This is optional and also will only be used for scheduling
        if there is a count field on a worker.

    :returns: A list where each element is a distinct subset of test_ids.
    """
    with open(worker_path, 'r') as worker_file:
        workers_desc = yaml.load(worker_file.read())
    worker_groups = []
    for worker in workers_desc:
        if isinstance(worker, dict) and 'worker' in worker.keys():
            if isinstance(worker['worker'], list):
                local_worker_list = selection.filter_tests(
                    worker['worker'], ids)

                if 'concurrency' in worker.keys() and worker[
                    'concurrency'] > 1:
                    partitioned_tests = partition_tests(
                        local_worker_list, worker['concurrency'], repository,
                        group_callback, randomize)
                    worker_groups.extend(partitioned_tests)
                else:
                    # If a worker partition is empty don't add it to the output
                    if local_worker_list:
                        worker_groups.append(local_worker_list)
            else:
                raise TypeError('The input yaml is the incorrect format')
        else:
            raise TypeError('The input yaml is the incorrect format')
    return worker_groups



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import operator

import mock
import six
import subunit
import testtools

from stestr import bisect_tests
from stestr.repository import abstract
from stestr.tests import base


class FakeTestRun(abstract.AbstractTestRun):
    def get_test(self):
        case = subunit.ByteStreamToStreamResult(io.BytesIO(self._content))

        def wrap_result(result):
            # Wrap in a router to mask out startTestRun/stopTestRun from the
            # ExtendedToStreamDecorator.
            result = testtools.StreamResultRouter(
                result, do_start_stop_run=False)
            # Wrap that in ExtendedToStreamDecorator to convert v1 calls to
            # StreamResult.
            return testtools.ExtendedToStreamDecorator(result)

        return testtools.DecorateTestCaseResult(
            case, wrap_result, operator.methodcaller('startTestRun'),
            operator.methodcaller('stopTestRun'))

    def get_id(self):
        return self.id


class FakeFailedTestRunNoTags(FakeTestRun):

    def __init__(self, failure=True):
        # Generate a subunit stream
        self.id = 2
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(test_id='test_a', test_status='inprogress')
        stream.status(test_id='test_a', test_status='success')
        stream.status(test_id='test_b', test_status='inprogress')
        stream.status(test_id='test_b', test_status='success')
        stream.status(test_id='test_c', test_status='inprogress')
        stream.status(test_id='test_c', test_status='fail')
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()


class FakeFailingWithTags(FakeTestRun):

    def __init__(self, failure=True):
        # Generate a subunit stream
        self.id = None
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(test_id='test_c', test_status='inprogress',
                      test_tags=['worker-0'])
        stream.status(test_id='test_c', test_status='fail',
                      test_tags=['worker-0'])
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()


class FakeNoFailing(FakeTestRun):

    def __init__(self, failure=True):
        # Generate a subunit stream
        stream_buf = io.BytesIO(six.binary_type(''.encode('utf-8')))
        self._content = stream_buf.getvalue()
        self.id = None


class FakeFailedTestRunWithTags(FakeTestRun):

    def __init__(self, failure=True):
        # Generate a subunit stream
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(test_id='test_a', test_status='inprogress',
                      test_tags=['worker-0'])
        stream.status(test_id='test_a', test_status='success',
                      test_tags=['worker-0'])
        stream.status(test_id='test_b', test_status='inprogress',
                      test_tags=['worker-1'])
        stream.status(test_id='test_b', test_status='success',
                      test_tags=['worker-1'])
        stream.status(test_id='test_c', test_status='inprogress',
                      test_tags=['worker-0'])
        stream.status(test_id='test_c', test_status='fail',
                      test_tags=['worker-0'])
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()
        self.id = 2


class FakeFailedMultiWorkerTestRunWithTags(FakeTestRun):

    def __init__(self, failure=True):
        # Generate a subunit stream
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(test_id='test_a', test_status='inprogress',
                      test_tags=['worker-0'])
        stream.status(test_id='test_a', test_status='success',
                      test_tags=['worker-0'])
        stream.status(test_id='test_b', test_status='inprogress',
                      test_tags=['worker-1'])
        stream.status(test_id='test_b', test_status='fail',
                      test_tags=['worker-1'])
        stream.status(test_id='test_c', test_status='inprogress',
                      test_tags=['worker-0'])
        stream.status(test_id='test_c', test_status='fail',
                      test_tags=['worker-0'])
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()
        self.id = 2


class TestBisectTests(base.TestCase):
    def setUp(self):
        super(TestBisectTests, self).setUp()
        self.repo_mock = mock.create_autospec(
            'stestr.repository.file.Repository')
        self.conf_mock = mock.create_autospec('stestr.config_file.TestrConf')
        self.run_func_mock = mock.MagicMock()
        self.latest_run_mock = mock.MagicMock()

    def test_bisect_no_failures_provided(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock,
            self.repo_mock)
        self.assertRaises(ValueError, bisector.bisect_tests, [])

    def test_prior_tests_invlaid_test_id(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock,
            self.repo_mock)
        run = FakeFailedTestRunNoTags()
        self.assertRaises(KeyError, bisector._prior_tests, run, 'bad_test_id')

    def test_get_prior_tests_no_tags(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock,
            self.repo_mock)
        run = FakeFailedTestRunNoTags()
        prior_tests = bisector._prior_tests(run, 'test_c')
        self.assertEqual(['test_a', 'test_b'], prior_tests)

    def test_get_prior_tests_with_tags(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock,
            self.repo_mock)
        run = FakeFailedTestRunWithTags()
        prior_tests = bisector._prior_tests(run, 'test_c')
        self.assertEqual(['test_a'], prior_tests)

    @mock.patch('stestr.output.output_table')
    def test_bisect_tests_isolated_failure(self, table_mock):
        run = FakeFailedTestRunWithTags()
        self.conf_mock.get_run_command = mock.MagicMock()

        def get_failures(*args, **kwargs):
            return FakeNoFailing()

        self.repo_mock.get_failing = get_failures
        bisector = bisect_tests.IsolationAnalyzer(
            run, self.conf_mock, self.run_func_mock, self.repo_mock)
        return_code = bisector.bisect_tests(['test_c'])
        expected_issue = [('failing test', 'caused by test'),
                          ('test_c', 'unknown - no conflicts')]
        table_mock.assert_called_once_with(expected_issue)
        self.assertEqual(3, return_code)

    @mock.patch('stestr.output.output_table')
    def test_bisect_tests_not_isolated_failure(self, table_mock):
        run = FakeFailedTestRunWithTags()
        self.conf_mock.get_run_command = mock.MagicMock()

        def get_failures(*args, **kwargs):
            return FakeFailingWithTags()

        self.repo_mock.get_failing = get_failures
        bisector = bisect_tests.IsolationAnalyzer(
            run, self.conf_mock, self.run_func_mock, self.repo_mock)
        return_code = bisector.bisect_tests(['test_c'])
        expected_issue = [('failing test', 'caused by test'),
                          ('test_c', 'test_a')]
        table_mock.assert_called_once_with(expected_issue)
        self.assertEqual(3, return_code)

    @mock.patch('stestr.output.output_table')
    def test_bisect_tests_not_isolated_multiworker_failures(self, table_mock):
        run = FakeFailedMultiWorkerTestRunWithTags()
        self.conf_mock.get_run_command = mock.MagicMock()

        def get_failures(*args, **kwargs):
            return FakeFailingWithTags()

        self.repo_mock.get_failing = get_failures
        bisector = bisect_tests.IsolationAnalyzer(
            run, self.conf_mock, self.run_func_mock, self.repo_mock)
        return_code = bisector.bisect_tests(['test_b', 'test_c'])
        expected_issue = [('failing test', 'caused by test'),
                          ('test_b', 'unknown - no conflicts'),
                          ('test_c', 'test_a')]
        table_mock.assert_called_once_with(expected_issue)
        self.assertEqual(3, return_code)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import print_function

import contextlib
import re
import sys


def filter_tests(filters, test_ids):
    """Filter test_ids by the test_filters.

    :param list filters: A list of regex filters to apply to the test_ids. The
        output will contain any test_ids which have a re.search() match for any
        of the regexes in this list. If this is None all test_ids will be
        returned
    :param list test_ids: A list of test_ids that will be filtered
    :return: A list of test ids.
    """
    if filters is None:
        return test_ids

    _filters = []
    for f in filters:
        if isinstance(f, str):
            try:
                _filters.append(re.compile(f))
            except re.error:
                print("Invalid regex: %s provided in filters" % f,
                      file=sys.stderr)
                sys.exit(5)
        else:
            _filters.append(f)

    def include(test_id):
        for pred in _filters:
            if pred.search(test_id):
                return True

    return list(filter(include, test_ids))


def black_reader(blacklist_file):
    with contextlib.closing(open(blacklist_file, 'r')) as black_file:
        regex_comment_lst = []  # tuple of (regex_compiled, msg, skipped_lst)
        for line in black_file:
            raw_line = line.strip()
            split_line = raw_line.split('#')
            # Before the # is the regex
            line_regex = split_line[0].strip()
            if len(split_line) > 1:
                # After the # is a comment
                comment = ''.join(split_line[1:]).strip()
            else:
                comment = 'Skipped because of regex %s:' % line_regex
            if not line_regex:
                continue
            try:
                regex_comment_lst.append((re.compile(line_regex), comment, []))
            except re.error:
                print("Invalid regex: %s in provided blacklist file" %
                      line_regex, file=sys.stderr)
                sys.exit(5)
    return regex_comment_lst


def _get_regex_from_whitelist_file(file_path):
    lines = []
    for line in open(file_path).read().splitlines():
        split_line = line.strip().split('#')
        # Before the # is the regex
        line_regex = split_line[0].strip()
        if line_regex:
            try:
                lines.append(re.compile(line_regex))
            except re.error:
                print("Invalid regex: %s in provided whitelist file" %
                      line_regex, file=sys.stderr)
                sys.exit(5)
    return lines


def construct_list(test_ids, blacklist_file=None, whitelist_file=None,
                   regexes=None, black_regex=None):
    """Filters the discovered test cases

    :param list test_ids: The set of test_ids to be filtered
    :param str blacklist_file: The path to a blacklist file
    :param str whitelist_file: The path to a whitelist file
    :param list regexes: A list of regex filters to apply to the test_ids. The
        output will contain any test_ids which have a re.search() match for any
        of the regexes in this list. If this is None all test_ids will be
        returned
    :param str black_regex:

    :return: iterable of strings. The strings are full
        test_ids
    :rtype: set
    """

    if not regexes:
        regexes = None  # handle the other false things

    white_re = None
    if whitelist_file:
        white_re = _get_regex_from_whitelist_file(whitelist_file)

    if not regexes and white_re:
        regexes = white_re
    elif regexes and white_re:
        regexes += white_re

    if blacklist_file:
        black_data = black_reader(blacklist_file)
    else:
        black_data = None

    if black_regex:
        msg = "Skipped because of regexp provided as a command line argument:"
        try:
            record = (re.compile(black_regex), msg, [])
        except re.error:
            print("Invalid regex: %s used for black_regex" % black_regex,
                  file=sys.stderr)
            sys.exit(5)
        if black_data:
            black_data.append(record)
        else:
            black_data = [record]

    list_of_test_cases = filter_tests(regexes, test_ids)
    set_of_test_cases = set(list_of_test_cases)

    if not black_data:
        return set_of_test_cases

    # NOTE(afazekas): We might use a faster logic when the
    # print option is not requested
    for (rex, msg, s_list) in black_data:
        for test_case in list_of_test_cases:
            if rex.search(test_case):
                # NOTE(mtreinish): In the case of overlapping regex the test
                # case might have already been removed from the set of tests
                if test_case in set_of_test_cases:
                    set_of_test_cases.remove(test_case)
                    s_list.append(test_case)

    return set_of_test_cases



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six

from stestr import output


class CallWhenProcFinishes(object):
    """Convert a process object to trigger a callback when returncode is set.

    This just wraps the entire object and when the returncode attribute access
    finds a set value, calls the callback.
    """

    def __init__(self, process, callback):
        """Adapt process

        :param process: A subprocess.Popen object.
        :param callback: The process to call when the process completes.
        """
        self._proc = process
        self._callback = callback
        self._done = False

    @property
    def stdin(self):
        return self._proc.stdin

    @property
    def stdout(self):
        return self._proc.stdout

    @property
    def stderr(self):
        return self._proc.stderr

    @property
    def returncode(self):
        result = self._proc.returncode
        if not self._done and result is not None:
            self._done = True
            self._callback()
        return result

    def wait(self):
        return self._proc.wait()


def _iter_internal_streams(input_streams, stream_type):
    streams = []
    for in_stream in input_streams:
        if in_stream[0] == stream_type:
            streams.append(in_stream[1])
    for stream_value in streams:
        if isinstance(stream_value, output.ReturnCodeToSubunit):
            if getattr(stream_value.source, 'detach', None):
                yield stream_value.source.detach()
            else:
                yield stream_value.source
        elif getattr(stream_value, 'read', None):
            yield stream_value
        else:
            yield six.BytesIO(stream_value)


def iter_streams(input_streams, stream_type):
    """Iterate over all the streams of type stream_type.

    :param stream_type: A simple string such as 'subunit' which matches
        one of the stream types defined for the cmd object this UI is
        being used with.
    :return: A generator of stream objects. stream objects have a read
        method and a close method which behave as for file objects.
    """
    for stream_spec in input_streams:
        _stream_spec = stream_spec[0]
        if '*' in _stream_spec or '?' in _stream_spec or '+' in _stream_spec:
            found = stream_type == _stream_spec[:-1]
        else:
            found = stream_type == _stream_spec
        if found:
            return _iter_internal_streams(input_streams, stream_type)
    raise KeyError(stream_type)


def cleanup_test_name(name, strip_tags=True, strip_scenarios=False):
    """Clean up the test name for display.

    By default we strip out the tags in the test because they don't help us
    in identifying the test that is run to it's result.

    Make it possible to strip out the testscenarios information (not to
    be confused with tempest scenarios) however that's often needed to
    identify generated negative tests.
    """
    if strip_tags:
        tags_start = name.find('[')
        tags_end = name.find(']')
        if tags_start > 0 and tags_end > tags_start:
            newname = name[:tags_start]
            newname += name[tags_end + 1:]
            name = newname

    if strip_scenarios:
        tags_start = name.find('(')
        tags_end = name.find(')')
        if tags_start > 0 and tags_end > tags_start:
            newname = name[:tags_start]
            newname += name[tags_end + 1:]
            name = newname

    return name



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re
import signal
import subprocess
import sys
import tempfile

import fixtures
import six
from subunit import v2

from stestr import results
from stestr import scheduler
from stestr import selection
from stestr import testlist


class TestProcessorFixture(fixtures.Fixture):
    """Write a temporary file to disk with test ids in it.

    The TestProcessorFixture is used to handle the lifecycle of running
    the subunit.run commands. A fixture is used for this class to handle
    the temporary list files creation.

    :param test_ids: The test_ids to use. May be None indicating that
        no ids are known and they should be discovered by listing or
        configuration if they must be known to run tests. Test ids are
        needed to run tests when filtering or partitioning is needed: if
        the run concurrency is > 1 partitioning is needed, and filtering is
        needed if the user has passed in filters.
    :param cmd_template: string to be used for the command that will be
        filled out with the IDFILE when it is created.
    :param listopt: Option to substitute into LISTOPT to cause test listing
                    to take place.
    :param idoption: Option to substitute into cmd when supplying any test ids.
    :param repository: The repository to query for test times, if needed.
    :param parallel: If not True, prohibit parallel use : used to implement
                     --parallel run recursively.
    :param listpath: The file listing path to use. If None, a unique path
                     is created.
    :param test_filters: An optional list of test filters to apply. Each
        filter should be a string suitable for passing to re.compile.
        Filters are applied using search() rather than match(), so if
        anchoring is needed it should be included in the regex.
        The test ids used for executing are the union of all the
        individual filters: to take the intersection instead, craft a
        single regex that matches all your criteria. Filters are
        automatically applied by run_tests(), or can be applied by calling
        filter_tests(test_ids).
    :param group_callback: If supplied, should be a function that accepts a
        test id and returns a group id. A group id is an arbitrary value
        used as a dictionary key in the scheduler. All test ids with the
        same group id are scheduled onto the same backend test process.
    :param bool serial: Run tests serially
    :param path worker_path: Optional path of a manual worker grouping file
        to use for the run
    :param int concurrency: How many processes to use. The default (0)
        autodetects your CPU count and uses that.
    :param path blacklist_file: Path to a blacklist file, this file contains a
        separate regex exclude on each newline.
    :param path whitelist_file: Path to a whitelist file, this file contains a
         separate regex on each newline.
    :param boolean randomize: Randomize the test order after they are
        partitioned into separate workers
    """

    def __init__(self, test_ids, cmd_template, listopt, idoption,
                 repository, parallel=True, listpath=None,
                 test_filters=None, group_callback=None, serial=False,
                 worker_path=None, concurrency=0, blacklist_file=None,
                 black_regex=None, whitelist_file=None, randomize=False):
        """Create a TestProcessorFixture."""

        self.test_ids = test_ids
        self.template = cmd_template
        self.listopt = listopt
        self.idoption = idoption
        self.repository = repository
        self.parallel = parallel
        if serial:
            self.parallel = False
        self._listpath = listpath
        self.test_filters = test_filters
        self._group_callback = group_callback
        self.worker_path = None
        self.worker_path = worker_path
        self.concurrency_value = concurrency
        self.blacklist_file = blacklist_file
        self.whitelist_file = whitelist_file
        self.black_regex = black_regex
        self.randomize = randomize

    def setUp(self):
        super(TestProcessorFixture, self).setUp()
        variable_regex = '\$(IDOPTION|IDFILE|IDLIST|LISTOPT)'
        variables = {}
        list_variables = {'LISTOPT': self.listopt}
        cmd = self.template
        default_idstr = None

        def list_subst(match):
            return list_variables.get(match.groups(1)[0], '')

        self.list_cmd = re.sub(variable_regex, list_subst, cmd)
        nonparallel = not self.parallel
        selection_logic = (self.test_filters or self.blacklist_file or
                           self.whitelist_file or self.black_regex)
        if nonparallel:
            self.concurrency = 1
        else:
            self.concurrency = None
            if self.concurrency_value:
                self.concurrency = int(self.concurrency_value)
            if not self.concurrency:
                self.concurrency = scheduler.local_concurrency()
            if not self.concurrency:
                self.concurrency = 1
        if self.test_ids is None:
            if self.concurrency == 1:
                if default_idstr:
                    self.test_ids = default_idstr.split()
            if self.concurrency != 1 or selection_logic or self.worker_path:
                # Have to be able to tell each worker what to run / filter
                # tests.
                self.test_ids = self.list_tests()
        if self.test_ids is None:
            # No test ids to supply to the program.
            self.list_file_name = None
            name = ''
            idlist = ''
        else:
            self.test_ids = selection.construct_list(
                self.test_ids, blacklist_file=self.blacklist_file,
                whitelist_file=self.whitelist_file,
                regexes=self.test_filters,
                black_regex=self.black_regex)
            name = self.make_listfile()
            variables['IDFILE'] = name
            idlist = ' '.join(self.test_ids)
        variables['IDLIST'] = idlist

        def subst(match):
            return variables.get(match.groups(1)[0], '')

        if self.test_ids is None:
            # No test ids, no id option.
            idoption = ''
        else:
            idoption = re.sub(variable_regex, subst, self.idoption)
            variables['IDOPTION'] = idoption
        self.cmd = re.sub(variable_regex, subst, cmd)

    def make_listfile(self):
        name = None
        try:
            if self._listpath:
                name = self._listpath
                stream = open(name, 'wb')
            else:
                fd, name = tempfile.mkstemp()
                stream = os.fdopen(fd, 'wb')
            with stream:
                self.list_file_name = name
                testlist.write_list(stream, self.test_ids)
        except Exception:
            if name:
                os.unlink(name)
            raise
        self.addCleanup(os.unlink, name)
        return name

    def _clear_SIGPIPE(self):
        """Clear SIGPIPE : child processes expect the default handler."""
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    def _start_process(self, cmd):
        # NOTE(claudiub): Windows does not support passing in a preexec_fn
        # argument.
        preexec_fn = None if sys.platform == 'win32' else self._clear_SIGPIPE
        return subprocess.Popen(cmd, shell=True,
                                stdout=subprocess.PIPE,
                                stdin=subprocess.PIPE,
                                preexec_fn=preexec_fn)

    def list_tests(self):
        """List the tests returned by list_cmd.

        :return: A list of test ids.
        """
        run_proc = self._start_process(self.list_cmd)
        out, err = run_proc.communicate()
        if run_proc.returncode != 0:
            sys.stdout.write("\n=========================\n"
                             "Failures during discovery"
                             "\n=========================\n")
            new_out = six.BytesIO()
            v2.ByteStreamToStreamResult(
                six.BytesIO(out), 'stdout').run(
                    results.CatFiles(new_out))
            out = new_out.getvalue()
            if out:
                if six.PY3:
                    sys.stdout.write(out.decode('utf8'))
                else:
                    sys.stdout.write(out)
            if err:
                if six.PY3:
                    sys.stdout.write(out.decode('utf8'))
                else:
                    sys.stderr.write(err)
            sys.stdout.write("\n" + "=" * 80 + "\n"
                             "The above traceback was encountered during "
                             "test discovery which imports all the found test"
                             " modules in the specified test_path.\n")
            exit(100)
        ids = testlist.parse_enumeration(out)
        return ids

    def run_tests(self):
        """Run the tests defined by the command

        :return: A list of spawned processes.
        """
        result = []
        test_ids = self.test_ids
        # Handle the single worker case (this is also run recursively per
        # worker in the parallel case)
        if self.concurrency == 1 and (test_ids is None or test_ids):
            run_proc = self._start_process(self.cmd)
            # Prevent processes stalling if they read from stdin; we could
            # pass this through in future, but there is no point doing that
            # until we have a working can-run-debugger-inline story.
            run_proc.stdin.close()
            return [run_proc]
        # If there is a worker path, use that to get worker groups
        elif self.worker_path:
            test_id_groups = scheduler.generate_worker_partitions(
                test_ids, self.worker_path, self.repository,
                self._group_callback, self.randomize)
        # If we have multiple workers partition the tests and recursively
        # create single worker TestProcessorFixtures for each worker
        else:
            test_id_groups = scheduler.partition_tests(test_ids,
                                                       self.concurrency,
                                                       self.repository,
                                                       self._group_callback)
        for test_ids in test_id_groups:
            if not test_ids:
                # No tests in this partition
                continue
            fixture = self.useFixture(
                TestProcessorFixture(test_ids,
                                     self.template, self.listopt,
                                     self.idoption, self.repository,
                                     parallel=False))
            result.extend(fixture.run_tests())
        return result
